# -*- coding: utf-8 -*-

from mptools.frameworks.py4web.controller import LocalsOnly

from .... import settings
from .common import pbfWebWrapper
from .callbacks import vtile_

@action(f'{settings.PATHROOT}/vtile/<xtile:int>/<ytile:int>/<zoom:int>', method=['GET'])
@action.uses(LocalsOnly())
@action.uses(pbfWebWrapper)
def vtile_xyz(xtile, ytile, zoom):
    return pbfWebWrapper(vtile_, x=xtile, y=ytile, z=zoom, source_name='osm')()

@action(f'{settings.PATHROOT}/vtile/<xtile:int>/<ytile:int>', method=['GET','POST'])
@action.uses(LocalsOnly())
@action.uses(pbfWebWrapper)
def vtile_xy(xtile, ytile):
    return pbfWebWrapper(vtile_, x=xtile, y=ytile, source_name='osm')()

@action(f'{settings.PATHROOT}/vtile', method=['GET','POST'])
@action.uses(LocalsOnly())
@action.uses(pbfWebWrapper)
def vtile():
    return pbfWebWrapper(vtile_, source_name='osm')()



# -*- coding: utf-8 -*-

from pydal.helpers.serializers import serializers
jsondumps = serializers.json

from pydal import geoPoint
from datetime import datetime

# import logging
# logger = logging.getLogger(__name__)

_get_coordinates = lambda lon, lat, height=None: (lon, lat,) if height is None else (lon, lat, height,)

class PanicError(Exception):
    """ """
    def __init__(self, message="It should never happen, why does it happen?"):
        super(PanicError, self).__init__(message)


class WTF(PanicError):
    """docstring for WTF."""


class __Base__(object):
    """docstring for __Base__."""

    def __init__(self, db, source_name):
        """
        db @DAL : The database in wich info, node, way_node and relation tables are defined.
        source_name @string : Source name
        """
        super(__Base__, self).__init__()
        self.db = db
        self.source_name = source_name


class BaseParser(__Base__):
    """docstring for BaseParser."""

    # def save_addon(self, key, properties):
    #     """ """
    #     ref_value = properties[key]
    #     rec = self.db.addon(
    #         source_name = self.source_name,
    #         ref_key = key,
    #         ref_value = ref_value
    #     )
    #     if rec is None:
    #         addon_id = db.addon.insert(
    #             properties = jsondumps(properties),
    #             ref_key = key,
    #             ref_value = ref_value,
    #             source_name = self.source_name,
    #         )
    #     else:
    #         raise NotImplementedError()
    #
    # def save_addons(self, key, addons):
    #     """ """

    def _save_info(self, rec_or_sid, gtype, tags=None, properties=None, attributes=None, merge=False):
        """
        rec    @row :
        sid @string : The entity identifier restricted to the data source environment.
        tags  @dict :
        """

        def insert_or_update(tags, properties, attributes, rec=None, **kw):
            if rec is None:
                id = self.db.info.insert(
                    tags = tags,
                    properties = properties,
                    attributes = attributes,
                **kw)
            elif not merge:
                rec.update_record(
                    tags = tags,
                    properties = properties,
                    attributes = attributes,
                **kw)
                id = rec.id
            else:
                rec.update_record(
                    tags = None if (rec.tags is None and tags is None) else dict(rec.tags or {}, **(tags or {})),
                    properties = None if (rec.properties is None and properties is None) else dict(rec.properties or {}, **(properties or {})),
                    attributes = None if (rec.attributes is None and attributes is None) else dict(rec.attributes or {}, **(attributes or {})),
                **kw)
                id = rec.id
            # logger.debug(f"Inserted info record with id: {id}")
            return id

        if not hasattr(rec_or_sid, "update_record"):
            # rec_or_sid is just an id
            sid = str(rec_or_sid)
            flt = dict(
                source_name = self.source_name,
                gtype = gtype,
                source_id = sid
            )
            rec = self.db.info(**flt)
            if not rec is None:
                if not attributes or rec.attributes.get('changeset')!=attributes.get('changeset'):
                    return insert_or_update(tags, properties, attributes, rec=rec, **flt)
                else:
                    return rec.id
            else:
                return insert_or_update(tags, properties, attributes, rec=rec, **flt)

        else:
            # rec_or_sid is a db record
            rec=rec_or_sid

            if not attributes or rec.attributes.get('changeset')!=attributes.get('changeset'):
                return insert_or_update(tags, properties, attributes, rec=rec)

    def _save_node(self, info_id, *coordinates):
        """
        info_id @integer :
        coordinates @tuple : lon, lat, height (optional)
        """
        point = geoPoint(*_get_coordinates(*coordinates))
        rec = self.db.node(info_id=info_id)

        if rec is None:
            return self.db.node.insert(info_id=info_id, geom=point)
        else:
            rec.update_record(geom=point)
            return rec.id

    def _save_way_node(self, info_id, node_id, sorting):
        rec = self.db.way_node(info_id=info_id, node_id=node_id)
        if rec is None:
            return self.db.way_node.insert(info_id=info_id, node_id=node_id, sorting=sorting)
        else:
            if rec.sorting!=sorting:
                rec.update_record(sorting=sorting)
                return rec.id

    def _save_relation(self, info_id, member_id, role):
        rec = self.db.relation(info_id=info_id, member_id=member_id)
        if rec is None:
            return self.db.relation.insert(info_id=info_id, member_id=member_id, role=role)
        else:
            rec.update_record(role=role)
            return rec.id


def normalize_tags_for_db(kwargs):
    _main = lambda c: (c[0], c[1] if not isinstance(c[1], str) else c[1].replace('"', "'"),)
    return dict(map(_main, kwargs.items()))


class BaseCopier(__Base__):
    """docstring for BaseCopier."""

    def _save_info(self, sid, gtype, tags=None, properties=None, attributes=None, **__):
        """
        sid @string : The entity identifier restricted to the data source environment.
        tags  @dict : OSM like tags
        properties @dict : Original raw properties
        attributes  @dict : OSM like attributes
        """

        _suid = dict(
            source_name = self.source_name,
            gtype = gtype,
            source_id = sid,
        )

        timestamp = datetime.now()

        flt = dict(
            _suid,
            tags = 'NULL' if tags is None else jsondumps(normalize_tags_for_db(tags)),
            properties = 'NULL' if properties is None else jsondumps(properties),
            attributes = 'NULL' if attributes is None else jsondumps(attributes),
            suid = self.db.info.suid.compute(_suid),
            created_on = timestamp,
            modified_on = timestamp
        )

        info_id = self._insinfo.writerow(flt)
        # TODO: forse
        # for
        # self.insfilter(dict(
        #
        # ))
        return info_id

    def _save_node(self, info_id, *coordinates):
        """
        info_id @integer :
        """
        point = geoPoint(*_get_coordinates(*coordinates))
        return self._insnodes.writerow(dict(
            info_id = info_id,
            geom = "SRID=4326;{}".format(point)
        ))

    def _save_way_node(self, info_id, node_id, sorting):
        return self._insways.writerow(dict(info_id=info_id, node_id=node_id, sorting=sorting))

    def _save_relation(self, info_id, member_id, role):
        # suid = self.db.relation.suid.compute(vars()),
        return self._insrelation.writerow(dict(
            info_id = info_id,
            member_id = member_id,
            role = role,
            # suid = suid
        ))



# -*- coding: utf-8 -*-

from ....common import db

# from py4vtile import settings as gpbfsettings
# if hasattr(gpbfsettings, "SHARE_DB") and gpbfsettings.SHARE_DB:
from py4vtile.pbfpp import common as pbfcommon
pbfcommon.db = db

from py4vtile.pbfpp import Prototizerpp as PbfPrototizer

pbfWebWrapper = PbfPrototizer()



# -*- coding: utf-8 -*-

from .sql import geoviews as views
from .. import settings

# import psycopg2.errors

DEFAULT_MATERIALIZED_VIEWS = [
    'sources',
    'segment_points',
    'spolys',
    'rpolys',
    'polys'
]

class BreakDatabaseIo(Exception):
    """ """

class Sqler(object):
    """docstring for Sqler."""

    def __init__(self):
        super(Sqler, self).__init__()
        from ..models import db
        self.db = db

    def __call__(self, query):
        """"""
        return self.db.executesql(query);

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if traceback:
            self.db.rollback()
            if isinstance(exc_value, BreakDatabaseIo):
                return True
        else:
            self.db.commit()

    def break_io(self, message=None):
        raise BreakDatabaseIo(message)

    def init_view(self, name, query, materialized=False):
        MATERIALIZED = ' MATERIALIZED' if materialized else ''
        _query = "DROP{MATERIALIZED} VIEW IF EXISTS {name} CASCADE;\n" if materialized else "DROP VIEW IF EXISTS {name} CASCADE;\n"
        _query += "CREATE{MATERIALIZED} VIEW {name} AS {query}"

        return self(_query.format(**vars()))


    def refresh_views(self):
        query = '; '.join(map(lambda vv: f"REFRESH MATERIALIZED VIEW {vv};", DEFAULT_MATERIALIZED_VIEWS))
        return self(query)


def setup():

    with Sqler() as sqler:

        # execute("""
        # CREATE SEQUENCE IF NOT EXISTS mysegseq;
        # ALTER SEQUENCE mysegseq RESTART WITH 1;
        # """)

        sqler("""
        CREATE SEQUENCE IF NOT EXISTS snodesseq;
        ALTER SEQUENCE snodesseq RESTART WITH 1;
        CREATE SEQUENCE IF NOT EXISTS spolyseq;
        ALTER SEQUENCE spolyseq RESTART WITH 1;
        CREATE SEQUENCE IF NOT EXISTS rpathseq;
        ALTER SEQUENCE rpathseq RESTART WITH 1;
        """)

        sqler.init_view('sources', views.SOURCES, materialized=('sources' in DEFAULT_MATERIALIZED_VIEWS))

        sqler.init_view('addresses', views.ADDRESSES, materialized=("addresses" in settings.MATERIALIZED_VIEWS))

        sqler.init_view('housenumbers', views.HOUSENUMBERS, materialized=("housenumbers" in settings.MATERIALIZED_VIEWS))

        sqler.init_view('points', views.POINTS, materialized=("points" in settings.MATERIALIZED_VIEWS))

        sqler.init_view('ways', views.WAYS, materialized=("ways" in settings.MATERIALIZED_VIEWS))

        sqler.init_view('graph', views.GRAPH, materialized=("graph" in settings.MATERIALIZED_VIEWS))

        sqler.init_view('segment_points', views.SPLITTEDSEGMENTSNODES,
            materialized = ('segment_points' in DEFAULT_MATERIALIZED_VIEWS)
        )

        sqler.init_view('spolys', views.SIMPLEPOLYGONS,
            materialized = ('spolys' in DEFAULT_MATERIALIZED_VIEWS)
        )

        sqler.init_view('rpath', views.REBUILDEDPOLYGONPATHS,
            materialized = ('rpath' in DEFAULT_MATERIALIZED_VIEWS)
        )

        sqler.init_view('rpolys', views.REBUILDEDPOLYGONS(rpath='rpath'),
            materialized = ('rpolys' in DEFAULT_MATERIALIZED_VIEWS)
        )

        sqler.init_view('polys', views.POLYS(spolys='spolys'),
            materialized = ('polys' in DEFAULT_MATERIALIZED_VIEWS)
        )

    return

# init_view('segment_points', Q4SPLITTEDSEGMENTSNODES, materialized=True)
#
# init_view('ways', Q4WAYS, materialized=True)
#
# init_view('graph', Q4GRAPH)
#
# execute("""
# CREATE SEQUENCE IF NOT EXISTS mypolyseq;
# SELECT setval('mypolyseq', (SELECT id+1 FROM ways ORDER BY id DESC LIMIT 1));
# """)
#
# init_view('spolys', Q4SIMPLEPOLYGONS,  materialized=True)
#
# init_view('rpath', Q4REBUILDEDPOLYGONPATHS, materialized=True)
#
# init_view('rpolys', Q4REBUILDEDPOLYGONS,  materialized=True)
#
# init_view('polys', '(SELECT * FROM spolys) UNION (SELECT * FROM rpolys);')



# -*- coding: utf-8 -*-

from py4web import action #, request, abort, redirect, URL, HTTP
from mptools.frameworks.py4web.controller import LocalsOnly

from .... import settings
from ...common import webWrapper
from .callbacks import fetch as fetch_
from .callbacks import guess_street as guess_street_
from .callbacks import fetcharound as fetcharound_

# WARNING!
# Subsequent controller are available from localhost only

@action(f'{settings.PATHROOT}/fetch', method=['GET', 'POST'])
@action.uses(LocalsOnly())
def fetch():
    return webWrapper(fetch_, source_name='osm')()

@action(f'{settings.PATHROOT}/fetcharound', method=['GET', 'POST'])
@action.uses(LocalsOnly())
def fetcharound():
    return webWrapper(fetcharound_, source_name='osm')()

@action(f'{settings.PATHROOT}/guess_street/<sugg>', method=['GET'])
@action.uses(LocalsOnly())
def guess_street(sugg):
    return webWrapper(guess_street_, sugg=sugg, comune='Genova', source='osm')()



# -*- coding: utf-8 -*-

import h3
import mercantile as mt

from shapely.geometry import Point
from shapely.ops import transform
from shapely.affinity import translate

from pyproj import Transformer, Proj

to3857 = Proj('epsg:3857')
to4326 = lambda x, y: to3857(x, y, inverse=True)

transformer = Transformer.from_crs("EPSG:4326", "EPSG:3857", always_xy=True)

def housenumber_components(hn):
    def loopOlettrs(ll):
        for l in ll:
            if l.isdigit():
                yield l
            else:
                break
    # components = {}
    number = ''.join(loopOlettrs(hn))
    letter, color = '', '',
    if hn.endswith('r'):
        color = 'r'
        letters = hn[len(number):-1].strip()
    elif hn.lower().endswith('rosso'):
        color = 'r'
        letters = hn[len(number):-len('rosso')].strip()
    else:
        letters = hn[len(number):]
    if letters:
        letter = letters
    return number and int(number), letter, color, hn,

def _get_buffered_bounds(minlon, minlat, maxlon, maxlat, zoom=18, classic=True):
    """
    minlon, minlat, maxlon, maxlat @float : Bbox limits;
    zoon @integer : Square tile zoom level or hexagonal tile resolution;
    classic @boolean : Whether to use classic square tiles or Uber H3 hexagonal ones.

    Returns the bbox limits that fits to the tiles touched by the bbox area introduced.
    In this way you are sure to fetch all the inolved points.
    """

    resolution = zoom
    if classic:
        # resolution = zoom
        ultile = mt.tile(minlon, maxlat, zoom)
        left, top = mt.ul(*ultile)
        rbtile = mt.tile(maxlon, minlat, zoom)
        brtile = lambda x, y, z: (x+1, y+1, z)
        right, bottom = mt.ul(*brtile(*rbtile))
        # get_tile = lambda lon, lat: mt.tile(lon, lat, zoom)
    else:
        # ultile = h3.geo_to_h3(maxlat, minlon, zoom)
        # resolution = min(12, zoom)

        ultile = h3.geo_to_h3(maxlat, minlon, resolution)
        ulboundary = h3.h3_to_geo_boundary(ultile)

        P1 = Point(ulboundary[0][::-1])
        P3 = Point(ulboundary[3][::-1])

        P1_3857 = transform(transformer.transform, P1)
        P3_3857 = transform(transformer.transform, P3)

        buffer = P1_3857.distance(P3_3857)

        ul_3857_ = transform(transformer.transform, Point((minlon, maxlat,)))

        ul_3857 = translate(ul_3857_, -buffer, buffer)

        br_3857_ = transform(transformer.transform, Point((maxlon, minlat,)))

        br_3857 = translate(br_3857_, buffer, -buffer)

        left, top = to3857(*ul_3857.coords[0], inverse=True)
        right, bottom = to3857(*br_3857.coords[0], inverse=True)

    return left, bottom, right, top, resolution,

def geomdbset(tab, minlon=None, minlat=None, maxlon=None, maxlat=None,
    source_name='__GENERIC__', tags=[], zoom_level=None, classic_tile=True, geom='geom'):
    """
    tab @pydal.table : DB table to query.
    minlon @float : Bounding box left limit longitude coordinate.
    minlat @float : Bounding box bottom limit latitude coordinate.
    maxlon @float : Bounding box right limit longitude coordinate.
    maxlat @float : Bounding box top limit latitude coordinate.
    source_name @text : Source name (ex.: osm).
    tags @list : List of dictionaries of tags to query for (ex.: [{'amenity': 'bar'}, ...]);
        Geometries resulting from query will must have at least all tags from any dictionary in the list.
    otags : Tags to query for.
        Geometries resulting from query will must be tagged as one of the passed tag.

    """
    basequery = (tab.source_name==source_name)

    if not any(map(lambda cc: cc is None, [minlon, minlat, maxlon, maxlat])):
        if not zoom_level is None:
            minlon, minlat, maxlon, maxlat, _ = _get_buffered_bounds(minlon, minlat, maxlon, maxlat, zoom=zoom_level, classic=classic_tile)
        basequery &= f"ST_Intersects({tab}.{geom}, ST_MakeEnvelope({minlon}, {minlat}, {maxlon}, {maxlat}, 4326))"

    if tags:
        basequery &= "("+" OR ".join([
            " AND ".join([
                f"({tab}.tags->>'{key}' = '{value}')" # .format(key=key, value=value, tab=tab) \
                    for key,value in tt.items()
                ]) for tt in tags
            ])+")"

    # logger.debug(db(basequery)._select())
    return tab._db(basequery)



# -*- coding: utf-8 -*-

import mercantile as mt
import h3
from shapely.geometry import Point
import pyproj

BASE_DIM = 155 # meters
MT_MAX_ZOOM = 30
H3_MAX_ZOOM = 15

def zoom2dims(zoom, lon, lat):
    tt = mt.tile(lon, lat, zoom)
    bb = mt.xy_bounds(tt)
    return bb.right-bb.left, bb.top-bb.bottom

def mt_tile_by_dim(lon, lat, bdim=BASE_DIM, asc=True, more=False):
    """
    lon  @float : Center longitude.
    lat  @float : Center latitude.
    bdim @float : Base tile dimension in meters.

    returns:
        tt @mercantile.tile : Tile nearest to desidered dimension coontaining the
                              point of given coordinates.
        zoom_level @integer : The tile zoom level.
    """
    max_area = bdim**2

    zooms = range(MT_MAX_ZOOM) if asc else reversed(range(MT_MAX_ZOOM))
    for zoom_level in zooms:
        tt = mt.tile(lon, lat, zoom_level)
        bb = mt.xy_bounds(tt)
        if (bb.right-bb.left)*(bb.top-bb.bottom) < max_area:
            if asc:
                break
        elif not asc:
            break
    return tt if not more else (tt, resolution,)

def h3_tile_by_dim(lon, lat, bdim=BASE_DIM, asc=True, more=False):

    osmproj = pyproj.Proj("epsg:3857")

    resolutions = range(H3_MAX_ZOOM) if asc else reversed(range(H3_MAX_ZOOM))
    for resolution in resolutions:
        tile = h3.geo_to_h3(lat, lon, resolution)
        polygon = h3.h3_to_geo_boundary(tile)

        x1, y1 = osmproj(*polygon[0])
        x2, y2 = osmproj(*polygon[1])

        dist = Point(x1, y1).distance(Point(x2, y2))

        if (asc and dist < bdim) or (not asc and dist > bdim):
            break

    return tile if not more else (tile, resolution,)

def tile_by_dim(lon, lat, bdim=BASE_DIM, asc=True, classic=False, more=False):
    if classic:
        return mt_tile_by_dim(lon, lat, bdim=bdim, asc=asc, more=more)
    else:
        return h3_tile_by_dim(lon, lat, bdim=bdim, asc=asc, more=more)



# -*- coding: utf-8 -*-

from ..tools import geomdbset

from ....setup.pgviews import Sqler

# WIP

def fetchtiled_(minlon, minlat, maxlon, maxlat, zoom=18, classic=False,
    source_name='__GENERIC__', tags=[]):
    """
    minlon, minlat, maxlon, maxlat @float : Bbox limits;
    zoon @integer : Square tile zoom level or hexagonal tile resolution;
    classic @boolean : Whether to use classic square tiles or Uber H3 hexagonal ones;
    source_name @string : Source name filter;
    tags @dict[] : Tags to be used as filter.

    Returns: Rows
    """

    tab = db.polys
    tiled_geom = 'centroid'

    if classic:
        tile_ = f"T_tile({tab}.{tiled_geom}, {zoom})"
        tilename_ = f"T_tilename({tile_})"
        get_poly_method = "T_bounds(buzz.tile)"
    else:
        tile_ = f"h3_geo_to_h3index({tab}.{tiled_geom}, {zoom})"
        tilename_ = tile_
        get_poly_method = "h3_h3index_to_geoboundary(buzz.tile)"

    dbset = geomdbset(tab, minlon=minlon, minlat=minlat, maxlon=maxlon, maxlat=maxlat,
        source_name=source_name, tags=tags, zoom_level=zoom, classic_tile=classic
    )

    tile = f"{tile_} as tile"
    tilename = f"{tilename_} as tilename"
    count = f"COUNT(0) as count"

    query = dbset._select(
        db.polys.id.min().with_alias('id'),
        tile,
        tilename,
        count,
        orderby = "tilename",
        groupby = "tilename, tile"
    )

    _view_name = 'buzz'

    with Sqler() as sqler:
        sqler.init_view(_view_name,
            f"SELECT {get_poly_method} as geom, * FROM ({query[:-1]}) as {_view_name}",
            materialized = False
        )

        db.define_table(_view_name,
            Field('geom', 'geometry()'),
            Field('tilename'),
            Field('count', 'integer'),
            Field('mrate', 'double'),
            Field.Virtual('feat_geometry', lambda row: wkt.loads(row[_view_name].geom)),
            migrate = False,
            redefine = True
        )

        result = db(db[_view_name]).select()
        sqler.break_io()

    return result



# -*- coding: utf-8 -*-

from ..models import db

from geojson import Point, Feature, FeatureCollection
from itertools import groupby
from .gjson import NodeParser, PolygonParser, WayParser
from .gjson import NodeCopier, PolygonCopier, WayCopier
from .osm import Parser, Copier
try:
    from ujson import loads as jsloads
except ImportError:
    from json import loads as jsloads

from tqdm import tqdm

from ..setup.pgviews import Sqler

def refresh(func):
    def wrapper(*args, **kwargs):
        res = func(*args, **kwargs)
        with Sqler() as sqler:
            sqler.refresh_views()
        return res
    return wrapper

@refresh
def feat(feature, source_name='__GENERIC__', copy=False, **kw):
    data = jsloads(feature) if isinstance(feature, str) else feature
    if not copy:
        if 'merge' in kw:
            merge = kw.pop('merge')
        else:
            merge = False
        if feature['geometry']['type']=='Point':
            node_parser = NodeParser(db, source_name, **kw)
            return node_parser.parse_feature(feature, merge=merge)
        elif feature['geometry']['type']=='Polygon':
            polygon_parser = PolygonParser(db, source_name, **kw)
            return polygon_parser.parse_feature(feature, merge=merge)
        else:
            NotImplementedError()
    else:
        if feature['geometry']['type']=='Point':
            node_parser = NodeCopier(db, source_name, **kw)
            return node_parser.parse_feature(feature)
        elif feature['geometry']['type']=='Polygon':
            polygon_parser = PolygonCopier(db, source_name, **kw)
            return polygon_parser.parse_feature(feature)
        else:
            NotImplementedError()
    raise NotImplementedError()

@refresh
def json(collection, source_name='__GENERIC__', copy=False, commit=True, **kw):
    data = jsloads(collection) if isinstance(collection, str) else collection
    grouping = lambda feat: feat['geometry']['type']
    data_by_type = groupby(sorted(data['features'], key=grouping), key=grouping)
    if copy:
        node_parser = NodeCopier(db, source_name, **kw)
        polygon_parser = PolygonCopier(db, source_name, **kw)
        linestring_parser = WayCopier(db, source_name, **kw)
    else:
        node_parser = NodeParser(db, source_name, **kw)
        polygon_parser = PolygonParser(db, source_name, **kw)
        linestring_parser = WayParser(db, source_name, **kw)

    for tt, features in data_by_type:
        if tt=='Point':
            node_parser.parse(features)
        elif tt=='Polygon':
            polygon_parser.parse(features)
        elif tt == 'LineString':
            linestring_parser.parse(features)
        else:
            raise NotImplementedError
    if commit:
        db.commit()

geojson = json

@refresh
def osm(nodes, ways, relations, copy=True, **kw):

    if copy:

        values = [
            db.info.suid.compute(
                {'source_name': 'osm', 'gtype': 'node', 'source_id': nn.id}
            ) for nn in nodes]

        values += [
            db.info.suid.compute(
                {'source_name': 'osm', 'gtype': 'way', 'source_id': nn.id}
            ) for nn in ways]

        values += [
            db.info.suid.compute(
                {'source_name': 'osm', 'gtype': 'relation', 'source_id': nn.id}
            ) for nn in relations]

        # data_to_insert = deepcopy(data)
        # data_to_update =

        res = db(
            db.info.suid.belongs(values)
        ).select(
            db.info.id,
            db.info.gtype,
            db.info.source_id
        ).group_by_value(db.info.gtype)

        already_in_db = {
            k: {int(row.source_id): row for row in rows} \
                for k,rows in res.items()
        }

        def is_not_in_db(k):

            try:
                myobjs = already_in_db[k]
            except KeyError:
                return lambda _: True
            else:
                return lambda nn: not str(nn.id) in myobjs

        Copier(db, source_name='OSM', **already_in_db).parse(
            nodes = filter(is_not_in_db('node'), nodes),
            ways = filter(is_not_in_db('way'), ways),
            relations = filter(is_not_in_db('relation'), relations)
        )

    else:

        Parser(db, source_name='OSM', **kw).parse(
            nodes = nodes,
            ways = ways,
            relations = relations
        )

@refresh
def idealista(cls, elementList, copy=False):
    def to_geojson():
        """ """
        for element in elementList:
            point = Point([element.pop('longitude'), element.pop('latitude')])
            yield Feature(geometry=point, properties=element, id=element['propertyCode'])

    return json(FeatureCollection(to_geojson()), source_name='idealista', copy=copy)



# -*- coding: utf-8 -*-

from py4geo.client.controllers.tools import _get_buffered_bounds, geomdbset

from ..setup.pgviews import Sqler
from ..common import db

class BaselTileViewsModel(Sqler):
    """ """

    GEOM_VIEW = 'gview'

    def __init__(self, minlon, minlat, maxlon, maxlat, zoom=18, classic=True,
        source_name='__GENERIC__', tags=[], geofield=db.points.geom):
        super(BaselTileViewsModel, self).__init__()
        self.GFIELD = geofield
        self.GTAB = geofield.table
        self.QUERIES = {}

        left, bottom, right, top, resolution = _get_buffered_bounds(
            minlon, minlat, maxlon, maxlat,
            zoom = zoom, classic = classic
        )

        if classic:
            tile_ = f"T_tile({self.GFIELD}, {resolution})"
            tilename_ = f"T_tilename({tile_})"
            # polygon_ = f"T_bounds({tile_})"
            self.get_poly_method = f"T_bounds({self.GEOM_VIEW}.tile)"
            self.get_area_method = f"ST_Area(ST_Transform({get_poly_method}, 3857))"

            # tilename_ = f"tilename(points.geom, {resolution})"
            # polygon_ = f"bounds_for_tile_indices(ST_Y(tile_indices_for_lonlat(points.geom, {resolution})), ST_X(tile_indices_for_lonlat(points.geom, {resolution})), {resolution})"
        else:
            tile_ = f"h3_geo_to_h3index({self.GFIELD}, {resolution})"
            tilename_ = tile_
            # polygon_ = f"h3_h3index_to_geoboundary(h3_geo_to_h3index(points.geom, {resolution}))"
            self.get_poly_method = f"h3_h3index_to_geoboundary({self.GEOM_VIEW}.tile)"
            self.get_area_method = f"h3_hexagon_area_km2({resolution})"

        self.tile = f"{tile_} as tile"
        self.tilename = f"{tilename_} as tilename"
        self.baseset = geomdbset(
            self.GTAB, left, bottom, right, top,
            source_name=source_name, tags=tags
        )

        self._queries_setup()

    def _queries_setup(self):
        """ """
        raise NotImplementedError()

    def _views_setup(self):
        for name, query in self.QUERIES.items():
            self.init_view(name, query)

    def __enter__(self):

        # Redefine this method

        # First running views setup method
        # self._views_setup()

        # And then setting up your views models
        # e.g. something like:
        # db.define_table(self.GEOM_VIEW,
        #     Field('geom', 'geometry()'),
        #     ...
        #     migrate = False,
        #     redefine = True
        # )

        raise NotImplementedError()

        # Last but not least returning self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.db.rollback()



# -*- coding: utf-8 -*-

from ..tools import geomdbset, housenumber_components
from ...models import db
from ....tools.tile import tilebbox

import mercantile as mt

def fetch_(minlon=None, minlat=None, maxlon=None, maxlat=None,
    source_name='__GENERIC__', tags=[], zoom_level=None, classic_tile=True,
    filter_geometries = None
):
    """ Returns a multi geometry type FeatureCollection accordingly to given tags
    and bounding box limits.
    minlon               @float :
    minlat               @float :
    maxlon               @float :
    maxlat               @float :
    source_name         @string :
    tags                @dict[] :
    zoom_level         @integer :
    classic_tile       @boolean :
    filter_geometries @callable : filter function for geometries.
    """

    def feats():
        for tab in filter(filter_geometries, (db.points, db.ways, db.polys,)):
            yield geomdbset(
                tab, minlon, minlat, maxlon, maxlat, source_name,
                tags=tags, zoom_level=zoom_level, classic_tile=classic_tile
            ).select()

    return map(lambda row: row.feature, chain(*feats()))

def fetch(minlon=None, minlat=None, maxlon=None, maxlat=None,
    source_name='__GENERIC__', tags=[], zoom_level=None, classic_tile=True,
    filter_geometries = None
):
    return {
        "type": "FeatureCollection",
        "features": list(fetch_(**vars()))
    }

def fetcharound(lon, lat, dist=200, bdim=None, buffer=0, source_name='__GENERIC__',
    tags=[], filter_geometries = None
):
    """ Returns a multi geometry type FeatureCollection accordingly to given tags
    around a center point.
    """
    _extra = {}
    if not bdim is None: _extra['bdim'] = bdim
    bbox = tilebbox(dist=dist, lon=lon, lat=lat, buffer=buffer, **_extra)
    # logger.debug(bbox)
    return fetch(
        minlon = bbox.minx,
        minlat = bbox.miny,
        maxlon = bbox.maxx,
        maxlat = bbox.maxy,
        source_name = source_name,
        tags = tags,
        filter_geometries = filter_geometries
    )

def vtile(x, y, z=18, source_name='__GENERIC__', tags=[], filter_geometries = None):
    """ """
    bounds = mt.bounds(x, y, z)
    feats_ = fetch_(
        minlon = bounds.west,
        minlat = bounds.south,
        maxlon = bounds.east,
        maxlat = bounds.north,
        source_name = source_name,
        tags = tags,
        filter_geometries = filter_geometries
    )
    return dict(features=feats_)

def guess_street(sugg, comune, source=None, limit=10):
    """ """

    if source=='disabled': return dict()

    words = filter(lambda e: e, re.split(";|,|\.|\n| |'", sugg))
    query = (db.housenumbers.housenumber!='') & \
        db.housenumbers.street.contains(list(words), all=True) & \
        db.housenumbers.city.contains(comune)

    if not source is None:
        query &= (db.housenumbers.source_name==source)

    housenumbers = "array_agg(housenumbers.housenumber)"

    res = db(query).select(
        housenumbers,
        db.housenumbers.street,
        db.housenumbers.city,
        db.housenumbers.street.lower(),
        groupby = db.housenumbers.street|db.housenumbers.city,
        orderby = db.housenumbers.street,
        limitby = (0,min((int(limit), 50,)),)
    ).group_by_value(db.housenumbers.street)

    return OrderedDict(
        sorted((k, OrderedDict(map(
            lambda tt: (tt[-1], {k: tt[i] for i,k in enumerate(['number', 'letter', 'color']) if tt[i]}),
            sorted(map(housenumber_components, [r[housenumbers] for r in v][0]))
        )),) \
        # sorted((k, map(housenumber_components, [r[housenumbers] for r in v][0]),) \
            for k,v in res.items()
        )
    )



# -*- coding: utf-8 -*-

import geojson
# import shapely.wkt
from hashids import Hashids
from pydal.objects import Table
from py4web import Field
from geomet import wkt
import mercantile as mc
import h3

def geojsonFeature(id, geometry, properties):
    return geojson.Feature(
        geometry = geometry,
        properties = properties,
        id = id
    )

def raise_error(err):
    raise err

hids = Hashids(min_length=1)

class PlanetTable(Table):
    """docstring for PlanetTable."""

    def __init__(self, *args, **kwargs):
        super(PlanetTable, self).__init__(*args, **kwargs)
        self._set_encoder()
        self._set_feature_co()

    def _set_encoder(self):
        self._extra = {'encoder': hids}
        self._extra["_decode"] = lambda encoded: self._extra['encoder'].decode(encoded) or \
            raise_error(Exception('String expected, "{}" found'.format(encoded)))
        self._extra["decode"] = lambda encoded: dict(zip(['src_id', 'id'], self._extra['encoder'].decode(encoded)))
        self._extra["get_by_hash"] = lambda encoded: self(**self._extra["decode"](encoded))
        # self._extra["fetch_info"] = lambda encoded: info(id=self._extra["_decode"](encoded)[1])

    def _set_alias(self, alias, fieldname):
        """ Sets an alias for the given field value """

        setattr(self, alias, Field.Virtual(alias,
            lambda row: row[self._tablename][fieldname]
        ))


    def _set_hashid(self, fieldname, first, *ofields):

        setattr(self, fieldname, Field.Virtual(fieldname,
            lambda row: self._extra['encoder'].encode(
                row[self._tablename][first],
                *map(lambda ff: row[self._tablename][ff], ofields)
            )
        ))

    def _set_feat_properties(self, **kwargs):

        def _props(row):
            properties = dict(
                row[self._tablename].properties or row[self._tablename].tags,
                id = row[self._tablename].hashid,
                # **{"_{}_".format(row[self._tablename].source_name): row[self._tablename].source_id}
            )
            properties.update({
                k: row[self._tablename][v] if not callable(v) else v(row[self._tablename])\
                    for k,v in kwargs.items()})
            return properties

        setattr(self, 'feat_properties', Field.Virtual('feat_properties', _props))

    def _set_geometry(self):

        if 'geom' in self.fields and self['geom'].type=='geometry()':
            self.feat_geometry = Field.Virtual('feat_geometry', lambda row: wkt.loads(row[self._tablename].geom))

    def _set_feature_co(self):
        self._set_hashid('hashid', 'src_id', 'id')
        self._set_feat_properties()
        self._set_geometry()

        if 'geom' in self.fields and self['geom'].type=='geometry()':

            self.feature = Field.Virtual('feature', lambda row: geojsonFeature(
                geometry = row[self._tablename].feat_geometry,
                properties = row[self._tablename].feat_properties,
                id = row[self._tablename].hashid
            ))

def get_tile(lon, lat, zoom, classic=True):
    if classic is True:
        return mc.tile(lon, lat, zoom)
    elif classic is False:
        return h3.geo_to_h3(lat, lon, resolution=zoom)

# def foo(row, zoom, classic=True):
#     import pdb; pdb.set_trace()

class PlanetPointTable(PlanetTable):
    """docstring for PlanetPointTable."""

    def __init__(self, *args, **kwargs):
        super(PlanetPointTable, self).__init__(*args, **kwargs)
        self._set_tile()

    def _set_tile(self):
        self.tile = Field.Method(
            'tile',
            lambda row, zoom, classic=True: get_tile(
                *wkt.loads(row[self._tablename].geom)["coordinates"],
                zoom = zoom,
                classic = classic
            )
        )

class PlanetGraphTable(PlanetTable):
    """docstring for PlanetGraphTable."""

    def _set_encoder(self):
        PlanetTable._set_encoder(self)
        self._extra['node_encoder'] = hids

    def _set_node_hashid(self, fieldname, first, *ofields):

        setattr(self, fieldname, Field.Virtual(fieldname,
            lambda row: self._extra['node_encoder'].encode(
                row[self._tablename][first],
                *map(lambda ff: row[self._tablename][ff], ofields)
            )
        ))

    def _set_feature_co(self):
        self._set_hashid('hashid', 'src_id', 'id')
        self._set_node_hashid('shashid', 'src_id', 'sinfo_id')
        self._set_node_hashid('thashid', 'src_id', 'tinfo_id')

        # self._set_alias("shid", 'shashid')
        # self._set_alias("thid", 'thashid')

        self._set_feat_properties(weight=lambda row: round(row.len, 3))
        if 'geom' in self.fields and self['geom'].type=='geometry()':

            self.feature = Field.Virtual('feature', lambda row: geojson.Feature(
                geometry = wkt.loads(row[self._tablename].geom),
                properties = row[self._tablename].feat_properties,
                id = row[self._tablename].hashid
            ))



# -*- coding: utf-8 -*-

import pyproj
import geojson

base_proj = "epsg:3857"
osmproj = pyproj.Proj(base_proj)

BASE_AREA_DIMENSION = 200
MIN_INFLUENCE_DISTANCE = 200

class GeoDict2(dict):
    """docstring for GeoDict2."""

    def __init__(self, geom={}, **kw):
        super(GeoDict2, self).__init__()
        self.__geom__ = {}

        for k,v in geom.items():
            try:
                self.__setgeom__(k, v)
            except WrongGeomType:
                pass

        for k,v in kw.items():
            try:
                self.__setgeom__(k, v)
            except WrongGeomType:
                pass
            else:
                kw.pop(k)

        self.update(kw)
        assert self.__geom__, "WARNING! No valid geom provided!"

    @classmethod
    def new(cls, **kw):
        new = cls(**kw)
        return new

    def __setgeom__(self, key, value):
        if not isinstance(value, Geom):
            raise WrongGeomType("WARNING! geom must be a 'Geom' object type.")

        assert (key=="local") or key.lower().startswith("epsg:") and key.split(":")[1].isdigit(), "Please respect the CRS name nomenclature!"
        self.__geom__[key.lower()] = value

    __getattr__ = dict.__getitem__

    @property
    def geom(self):
        return self.__geom__

    @property
    def id(self):
        return self.__id__

    @property
    def properties(self):
        return {k: v for k,v in self.items()}

    def get_feature(self, crs="epsg:4326", **kw):
        try:
            feat = self.geom[crs].feature
        except KeyError:
            myproj = pyproj.Proj(crs)
            e, n = myproj(*self.geom["epsg:4326"].xy)
            geom = Point(e, n, type="EPSG", properties={'code': crs.split(":")[1]})
            feat = geom.feature
        return geojson.Feature(geometry=feat.geometry, id=self.id, properties=dict(self.properties, **kw))

    def coordinates(self, crs=base_proj):
        return self.geom[crs].coordinates

    @classmethod
    def OSMSerialfactory(cls, OSMNodes, **kw):
        for OSMNode in OSMNodes:
            yield cls.OSMfactory(OSMNode, **kw)


class Bbox(GeoDict2):
    """ """
    buffer = None

    def get_shape(self, crs=base_proj):
        """ Returns the _external polygon_ as a shapely object """
        coordinates = self.geom[crs].coordinates
        return shPolygon(coordinates[0])

    def __contains__(self, lonlat):
        crs = 'epsg:4326'
        bbox_as_shape = self.get_shape(crs)
        node_as_shape = Node.factory(*lonlat).get_shape(crs)
        return bbox_as_shape.contains(node_as_shape)


    @classmethod
    def factory(cls, minlon, minlat, maxlon, maxlat, crss=(base_proj,)):

        _crss = (crs for crs in crss if crs!="epsg:4326")

        _ring = (minlon, minlat,), \
            (maxlon, minlat,), \
            (maxlon, maxlat,), \
            (minlon, maxlat,), \
            (minlon, minlat,),

        geom = dict(Polygon.projfactory((_ring,), crs=crs) for crs in _crss)
        geom["epsg:4326"] = Polygon((_ring,))

        return cls(geom = geom)

    @classmethod
    def bufferfactory(cls, lon, lat,
        buffer = BASE_AREA_DIMENSION+MIN_INFLUENCE_DISTANCE,
        crss = (base_proj,)
    ):
        """
        lon    @float : Longitude
        lat    @float : Latitude
        buffer @float : the buffer length in meters
        crss    @list : A list of valid crs
        """
        _crss = set([crs for crs in crss if crs!="epsg:4326"] + [base_proj])

        def _ring(crs):
            myproj = pyproj.Proj(crs)
            e, n = myproj(lon, lat)
            mine = e - buffer
            maxe = e + buffer
            minn = n - buffer
            maxn = n + buffer
            return crs, ((mine, minn,), \
                (maxe, minn,), \
                (maxe, maxn,), \
                (mine, maxn,),
                (mine, minn,),)

        # import pdb; pdb.set_trace()
        geom = dict([Polygon.factory((_r,), crs=_crs) for _crs,_r in ((a,b) for a,b in (_ring(crs) for crs in _crss))])

        myproj = pyproj.Proj(_crs)
        minlon, minlat = myproj(*map(min, zip(*_r)), inverse=True)
        maxlon, maxlat = myproj(*map(max, zip(*_r)), inverse=True)
        _ring = (minlon, minlat,), \
            (maxlon, minlat,), \
            (maxlon, maxlat,), \
            (minlon, maxlat,), \
            (minlon, minlat,),
        geom["epsg:4326"] = Polygon((_ring,))
        bbox_new = cls.new(geom=geom)
        bbox_new.buffer = buffer
        return bbox_new

    def min(self, crs="epsg:4326"):
        return tuple(map(min, zip(*self.geom[crs].coordinates[0])))

    def max(self, crs="epsg:4326"):
        return tuple(map(max, zip(*self.geom[crs].coordinates[0])))

    def center_and_max_dist(self, crs="epsg:3857"):
        mine, minn = self.min(crs)
        maxe, maxn = self.max(crs)
        dmax = max([maxe-mine, maxn-minn])/2
        center = Node.enfactory(mine+(maxe-mine)/2, minn+(maxn-minn)/2)
        return center, dmax

    @property
    def minlon(self):
        return self.min(crs="epsg:4326")[0]

    @property
    def minlat(self):
        return self.min(crs="epsg:4326")[1]

    @property
    def maxlon(self):
        return self.max(crs="epsg:4326")[0]

    @property
    def maxlat(self):
        return self.max(crs="epsg:4326")[1]

    def _lbrt(self, crs="epsg:4326"):
        """ """
        minx, miny = self.min(crs=crs)
        maxx, maxy = self.max(crs=crs)
        return minx, miny, maxx, maxy

    @property
    def lbrt(self):
        return self._lbrt(crs="epsg:4326")

    def _nsew(self, crs="epsg:4326"):
        minx, miny, maxx, maxy = self._lbrt(crs=crs)
        return maxy, miny, maxx, minx

    @property
    def nsew(self):
        return self._nsew(crs="epsg:4326")

    @property
    def osmbbox(self):
        values = self.lbrt
        keys = ("w", "s", "e", "n",)
        return dict(zip(keys, map(str, values)))

    def split(self, n=10):
        incr = self.buffer * 2 / float(n)
        first = incr/2.
        coordinates = self.coordinates()[0][0]
        cellcoords = []
        side = first
        for i in range(n):
            for j in range(n):
                est, nord = osmproj( coordinates[0] + first + (incr*i), coordinates[1] + first + (incr*j) , inverse=True)
                yield Bbox.bufferfactory(est, nord, side)

class Geom(object):

    known_crs = {
        # * http://spatialreference.org/ref/sr-org/7483/
        # Projection used in many popular web mapping applications (Google/Bing/OpenStreetMap/etc).
        # Sometimes known as EPSG:900913.
        "epsg:3857": {"link": {
            "href": "http://spatialreference.org/ref/sr-org/7483/proj4/",
            "type": "proj4"
        }}
    }

    def __init__(self, feature):
        validation = geojson.is_valid(feature)
        assert validation['valid']=='yes', validation['message']
        super(Geom, self).__init__()
        self.feature = feature

    __repr__ = lambda self: self.feature.__repr__()

    __str__ = __repr__

    @property
    def geometry(self):
        return self.feature["geometry"]

    geom = geometry

    def __call__(self):
        return self.geom

    @property
    def crs(self):
        return self.feature.get("crs")

    def extract(self):
        """ Extracts basic components """
        return self.feature["crs"], self.feature["geometry"]

    @property
    def coordinates(self):
        return tuple(self.feature["geometry"]["coordinates"])


class Polygon(Geom):

    def __init__(self, rings, properties=None, type="link"):
        """
        rings @list/@tuple : List of rings part of the polygon;
            for Polygons with multiple rings, the first must be the exterior
            ring and any others must be interior rings or holes.
        properties @dict : CRS properties
        type @str : Type of CRS propeties informations
        """
        kw = {} if properties is None else {"crs": {"type": type, "properties": properties}}
        self.feature = geojson.Feature(
            geometry = geojson.Polygon(rings),
            **kw
        )

    @classmethod
    def factory(cls, rings, crs=base_proj):
        """ """
        _type, props = list(cls.known_crs[crs].items())[0]
        return crs,  cls(rings, properties=props, type=_type)

    @classmethod
    def projfactory(cls, _rings, crs=base_proj, **kw):
        """
        Build projected polygon starting from rings or sequences of vertices
        in longitude and latitude.
        _rings @list : List of longitude and latitude coordinates;
        crs  @string : Name of the desidered crs
        """
        custom_proj = pyproj.Proj(crs)
        try:
            _type = kw["type"]
            props = kw["properties"]
        except KeyError:
            _type, props = list(cls.known_crs[crs].items())[0]

        rings = tuple(map(
            lambda ring: tuple(map(
                lambda coords: custom_proj(*coords),
                ring
            )),
            _rings
        ))
        return crs, cls(rings, properties=props, type=_type)

class Bbox(GeoDict2):
    """ """
    buffer = None

    def get_shape(self, crs=base_proj):
        """ Returns the _external polygon_ as a shapely object """
        coordinates = self.geom[crs].coordinates
        return shPolygon(coordinates[0])

    def __contains__(self, lonlat):
        crs = 'epsg:4326'
        bbox_as_shape = self.get_shape(crs)
        node_as_shape = Node.factory(*lonlat).get_shape(crs)
        return bbox_as_shape.contains(node_as_shape)


    @classmethod
    def factory(cls, minlon, minlat, maxlon, maxlat, crss=(base_proj,)):

        _crss = (crs for crs in crss if crs!="epsg:4326")

        _ring = (minlon, minlat,), \
            (maxlon, minlat,), \
            (maxlon, maxlat,), \
            (minlon, maxlat,), \
            (minlon, minlat,),

        geom = dict(Polygon.projfactory((_ring,), crs=crs) for crs in _crss)
        geom["epsg:4326"] = Polygon((_ring,))

        return cls(geom = geom)

    @classmethod
    def bufferfactory(cls, lon, lat,
        buffer = BASE_AREA_DIMENSION+MIN_INFLUENCE_DISTANCE,
        crss = (base_proj,)
    ):
        """
        lon    @float : Longitude
        lat    @float : Latitude
        buffer @float : the buffer length in meters
        crss    @list : A list of valid crs
        """
        _crss = set([crs for crs in crss if crs!="epsg:4326"] + [base_proj])

        def _ring(crs):
            myproj = pyproj.Proj(crs)
            e, n = myproj(lon, lat)
            mine = e - buffer
            maxe = e + buffer
            minn = n - buffer
            maxn = n + buffer
            return crs, ((mine, minn,), \
                (maxe, minn,), \
                (maxe, maxn,), \
                (mine, maxn,),
                (mine, minn,),)

        # import pdb; pdb.set_trace()
        geom = dict([Polygon.factory((_r,), crs=_crs) for _crs,_r in ((a,b) for a,b in (_ring(crs) for crs in _crss))])

        myproj = pyproj.Proj(base_proj)
        minlon, minlat = myproj(*map(min, zip(*_ring(base_proj)[1])), inverse=True)
        maxlon, maxlat = myproj(*map(max, zip(*_ring(base_proj)[1])), inverse=True)
        _ring = (minlon, minlat,), \
            (maxlon, minlat,), \
            (maxlon, maxlat,), \
            (minlon, maxlat,), \
            (minlon, minlat,),
        geom["epsg:4326"] = Polygon((_ring,))
        bbox_new = cls.new(geom=geom)
        bbox_new.buffer = buffer
        return bbox_new

    def min(self, crs="epsg:4326"):
        return tuple(map(min, zip(*self.geom[crs].coordinates[0])))

    def max(self, crs="epsg:4326"):
        return tuple(map(max, zip(*self.geom[crs].coordinates[0])))

    def center_and_max_dist(self, crs="epsg:3857"):
        mine, minn = self.min(crs)
        maxe, maxn = self.max(crs)
        dmax = max([maxe-mine, maxn-minn])/2
        center = Node.enfactory(mine+(maxe-mine)/2, minn+(maxn-minn)/2)
        return center, dmax

    @property
    def minlon(self):
        return self.min(crs="epsg:4326")[0]

    @property
    def minlat(self):
        return self.min(crs="epsg:4326")[1]

    @property
    def maxlon(self):
        return self.max(crs="epsg:4326")[0]

    @property
    def maxlat(self):
        return self.max(crs="epsg:4326")[1]

    def _lbrt(self, crs="epsg:4326"):
        """ """
        minx, miny = self.min(crs=crs)
        maxx, maxy = self.max(crs=crs)
        return minx, miny, maxx, maxy

    @property
    def lbrt(self):
        return self._lbrt(crs="epsg:4326")

    def _nsew(self, crs="epsg:4326"):
        minx, miny, maxx, maxy = self._lbrt(crs=crs)
        return maxy, miny, maxx, minx

    @property
    def nsew(self):
        return self._nsew(crs="epsg:4326")

    @property
    def osmbbox(self):
        values = self.lbrt
        keys = ("w", "s", "e", "n",)
        return dict(zip(keys, map(str, values)))

    def split(self, n=10):
        incr = self.buffer * 2 / float(n)
        first = incr/2.
        coordinates = self.coordinates()[0][0]
        cellcoords = []
        side = first
        for i in range(n):
            for j in range(n):
                est, nord = osmproj( coordinates[0] + first + (incr*i), coordinates[1] + first + (incr*j) , inverse=True)
                yield Bbox.bufferfactory(est, nord, side)



# -*- coding: utf-8 -*-

import csv, io
# from swissknife.log import timeLoggerDecorator

class BulkCopyer(object):
    """
    Courtesy of: https://www.citusdata.com/blog/2017/11/08/faster-bulk-loading-in-postgresql-with-copy/
    TODO: https://stackoverflow.com/a/8150329/1039510
    https://stackoverflow.com/a/9166750/1039510
    """

    # def __nextid(self):
    #     """ [1] TODO:
    #     https://github.com/web2py/pydal/blob/60e97e7cfd1da98f3cf38b2023965226d42e5e5b/pydal/adapters/postgres.py#L179
    #     """
    #     last_row = self.table._db(self.table).select(self.table.id, orderby=~self.table.id, limitby=(0,1,)).first()
    #     import pdb; pdb.set_trace()
    #     return 0 if last_row is None else last_row.id
    #     return self.db.executesql("select nextval('{}_id_seq')".format(self.table_name))[0][0]-1

    def __init__(self, table):
        """
        table @DAL.Table : "db.graph_path"
        """
        super(BulkCopyer, self).__init__()

        self.db = table._db
        self.adapter = table._db._adapter
        self.table_name = table._tablename
        self.table = table
        csv.register_dialect('custom', delimiter='\t', quotechar="", quoting=csv.QUOTE_NONE)
        self.csv = io.StringIO()
        self.writer = csv.DictWriter(self.csv, fieldnames=table.fields(), dialect='custom')

    def __enter__(self):
        # self.maxid = self.__nextid()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        # self.csv.seek(0)

        if traceback is None:
            # foo = {}
            # if not current.plugins.planetstore.log_level is None:
            #     foo['level'] = current.plugins.planetstore.log_level

            # with timeLoggerDecorator("Coping to {}".format(self.table_name), **foo):
            content = self.csv.getvalue()
            if content:
                try:
                    self.adapter.cursor.copy_from(io.StringIO(content), self.table_name, null='NULL')
                except Exception as err:
                    self.db.rollback()
                    with open("/tmp/{}".format(self.table_name), "w") as foo:
                        foo.write(self.csv.getvalue())
                    raise
                else:
                #     maxid = self.db(self.table).select(self.table.id.max().with_alias('maxid')).first().maxid
                #     nextid = self.__nextid()
                #     if nextid<=maxcid:
                #         self.db.executesql(syncquery(self.table_name))
                    self.db.commit()
        else:
            raise

    def writerow(self, d):
        """ -> [1] """
        # self.maxid += 1
        d['id'] = self.db.executesql("select nextval('{}_id_seq')".format(self.table_name))[0][0]
        self.writer.writerow(d)
        return d['id']



# -*- coding: utf-8 -*-

"""
Still experimental but useful
"""

from .tools import Syncher
from ..tools.tile import Bbox
from .optutils.base import Turbo
import json, geojson
import osm2geojson
from supermercado.burntiles import burn
import mercantile
from tqdm import tqdm

import overpy
# https://overpass-api.de/api/interpreter
# https://overpass.osm.ch/api/interpreter
overpy.Overpass.default_url = "https://overpass.kumi.systems/api/interpreter"

# minlon, minlat, maxlon, maxlat = 6.43798828125, 36.50963615733049, 18.6328125, 47.18971246448421,

edu_query = [
    [{"k": 'amenity', "v": 'college'}],
    [{"k": 'amenity', "v": 'driving_school'}],
    [{"k": 'amenity', "v": 'kindergarten'}],
    [{"k": 'amenity', "v": 'language_school'}],
    [{"k": 'amenity', "v": 'library'}],
    [{"k": 'amenity', "v": 'toy_library'}],
    [{"k": 'amenity', "v": 'music_school'}],
    [{"k": 'amenity', "v": 'school'}],
    [{"k": 'amenity', "v": 'university'}],
]

transp_query = [
    [{"k": "highway", "v": "bus_stop"}],
    [{"k": "public_transport", "v": "platform"}],
    [{"k": "amenity", "v": "parking"}],
    [{"k": "amenity", "v": "parking_space"}],
    [{"k": "amenity", "v": "fuel"}],
    [{"k": "amenity", "v": "bus_station"}]
]

def get_polys(featureCollection):
    for feature in featureCollection['features']:
        if feature['geometry']['type'] == 'MultiPolygon':
            for poly in feature['geometry']['coordinates']:
                xfeature = { "type": "Feature", "properties": {}, "geometry": { "type": "Polygon" } }
                xfeature['geometry']['coordinates'] = poly
                yield xfeature
        elif feature['geometry']['type'] == 'Polygon':
            yield feature

def tile2poly(xyz):
    x, y, z = xyz
    bounds = mercantile.bounds(x, y, z)
    bbox = Bbox(bounds.west, bounds.south, bounds.east, bounds.north)
    poly = bbox.as_gj_polygon
    return poly


def smart_osm_fetch(nome, zoom=14, admin_level=8):

    def flt():
        yield {
            # "bbox": Bbox(minlon, minlat, maxlon, maxlat).osm,
            "query": [[
                {"k": "boundary", "v": "administrative"},
                {"k": "admin_level", "v": f"{admin_level:d}"},
                {"k":"name", "v": f"{nome}"},
                {"k": "wikipedia", "regv": "it:"}
            ]],
            "gtypes": ['relation']
        }

    query = Turbo.build_query(flt)
    turbo = Turbo()
    data_, _ = turbo.__raw_call__(query.encode())
    data = json.loads(data_)
    fc = osm2geojson.json2geojson(data)

    polys = list(get_polys(fc))

    assert len(polys)>0

    tiles = burn(polys, zoom)

    coll = geojson.FeatureCollection(list(map(
        lambda gg: geojson.Feature(geometry=gg),
        map(tile2poly, tiles)
    )))

    for tile in tqdm(tiles):

        download = Syncher(*tile, query=edu_query+transp_query)
        download()



# -*- coding: utf-8 -*-

from py4web import Field
from itertools import chain

from ..common import T, db
from .table import PlanetTable, PlanetPointTable, PlanetGraphTable

db._check_reserved = False

addrsfields = lambda *of: (Field("source_name"), Field("city"), Field("street"),) + of

db.define_table("addresses",
    *addrsfields(),
    # Field("codvia"),
    migrate = False
)

db.define_table("housenumbers",
    *addrsfields(
        Field('housenumber')
    ),
    migrate = False
)

# db.define_table("addons",
#     Field("src_id"),
#     Field("source_name"),
#     Field("source_id"),
#     Field("properties", "json"),
#     singular = T("addon"), plural = T("addons"),
#     table_class = PlanetTable,
#     migrate = False
# )

db.define_table("points",
    Field("node_id", "bigint"),
    Field("src_id"),
    Field("source_name"),
    Field("source_id"),
    Field("geom", "geometry()"),
    Field("tags", "json"),
    Field("properties", "json"),
    Field("crds", "json"),
    singular = T("point"), plural = T("points"),
    table_class = PlanetPointTable,
    migrate = False
)

polyssfields = lambda *of: (
    Field("src_id"),
    Field("source_name"),
    Field("source_id"),
    Field("geom", "geometry()"),
    # Field("centroid", "geometry()"),
    Field("tags", "json"),
    Field("properties", "json"),
) + of

# db.define_table("ways",
#     *polyssfields(),
#     # Field("centroid", "geometry()"),
#     table_class = PlanetTable,
#     migrate = False
# )

db.define_table("polys",
    *polyssfields(),
    Field("centroid", "geometry()"),
    table_class = PlanetTable,
    migrate = False
)

# db.define_table("mpolys",
#     *polyssfields(),
#     Field("centroid", "geometry()"),
#     table_class = PlanetTable,
#     migrate = False
# )

# db.define_table("splitted_polys",
#     *polyssfields(),
#     table_class = PlanetTable,
#     migrate = False,
#     rname = "_splitted_polys"
# )

db.define_table("graph",
    Field("src_id"),
    Field("source_id", readable=False),
    Field("source_name"),
    Field("geom", "geometry()", readable=False),
    # Field("ssource_id", readable=False),
    Field("sinfo_id", "bigint", readable=False),
    Field("snode_id", "bigint", readable=False),
    Field("stags", "json", readable=False),
    # Field("tsource_id", readable=False),
    Field("tinfo_id", "bigint", readable=False),
    Field("tnode_id", "bigint", readable=False),
    Field("ttags", "json", readable=False),
    Field("tags", "json"),
    Field("properties", "json"),
    # Field("crds", "json", readable=False),
    Field("len", "double"),
    # Field("crs"),
    # Field("snid", "reference node", label='Source node id'),
    # Field("tnid", "reference node", label='Target node id'),
    table_class = PlanetGraphTable,
    migrate = False
)



# -*- coding: utf-8 -*-

from ..common import db
from .tools import Syncher, get_uri

from py4web import Field
import datetime
import mercantile as mc

now = lambda : datetime.datetime.utcnow()

db.define_table("tracked_tile",
    Field('xtile', 'integer', required=True, notnull=True),
    Field('ytile', 'integer', required=True, notnull=True),
    Field('zoom', 'integer', required=True, notnull=True, default=18),
    Field('uri', required=False, unique=True, notnull=True,
        compute = lambda row: get_uri(row['xtile'], row['ytile'], row['zoom'])
    ),
    # Field('filled', 'boolean', default=False),
    Field('created_on', 'datetime',
        # required = True,
        notnull = True,
        default = now,
        writable = False, readable = True
    ),
    Field('modified_on', 'datetime',
        # required = True, # notnull=True,
        update = now,
        default = now,
        # compute = lambda _=None: now(),
        writable = False, readable = True
    ),
    Field("is_active", "boolean", default=False, readable=False, writable=False),
    # Field('task_id', "reference scheduler_task", notnull=True, requires=None),
    Field.Virtual('feature', lambda row: mc.feature(
        mc.quadkey_to_tile(mc.quadkey(row.tracked_tile.xtile, row.tracked_tile.ytile, row.tracked_tile.zoom)),
        fid = row.tracked_tile.uri,
        props = {'created': row.tracked_tile.created_on, 'updated': row.tracked_tile.modified_on}
    )),
    # Field.Virtual('last_update', lambda row: get_last_update(row.tile.uri))
)


class OsmGetter(object):
    """docstring for OsmGetter."""

    def __init__(self):
        super(OsmGetter, self).__init__()
        self.rec = db.tracked_tile(is_active=False)

    def __enter__(self):
        self.rec.update_record(is_active=True)
        run = Syncher(self.rec.xtile, self.rec.ytile, self.rec.zoom, self.rec.uri)
        run(pgcopy=False)

    def __exit__(self, exc_type, exc_value, traceback):
        if not traceback is None:
            self.rec.update_record(is_active=False)

class OsmUpdater(object):
    """ TODO """



# -*- coding: utf-8 -*-

from .common import db, T

from py4web import Field
import datetime

now = lambda : datetime.datetime.utcnow()

db._check_reserved = False

class Info(Field):
    """ """

    def __init__(self, fieldname='info_id', type='reference info', *args, **kwargs):
        super(Info, self).__init__(fieldname, type, *args, **kwargs)

db.define_table("info",
    Field("source_name", required=True, notnull=True),
    Field("source_id", required=True, notnull=True),
    Field("properties", "json", required=True, notnull=False,
        label = T("External properties"),
        comment = T("External properties as they come from source")
    ),
    Field('created_on', 'datetime',
        default = now,
        writable=False, readable=False,
        label = T('Created On')
    ),
    Field('modified_on', 'datetime',
        update=now, default=now,
        writable=False, readable=False,
        label=T('Modified On')
    ),
    Field("tags", "json", label=T('OSM tags'), required=False, notnull=False),
    Field("attributes", "json", rname="attrs", label=T('OSM meta attributes'), required=False, notnull=False),
    Field("gtype", label=T('Geometry type'), # notnull=True,
        writable=False, readable=True
    ),
    Field("suid", unique=True, required=True, notnull=True,
        compute = lambda r: "{source_name}-{gtype}-{source_id}".format(**r)
    ),
    Field("is_active", "boolean", default=True, update=True)
)

db.define_table("node",
    Info(),
    Field("geom", "geometry()")
)

db.define_table("way_node",
    Info(),
    Field("node_id", "reference node", notnull=True),
    Field("sorting", "integer", default=0, notnull=True, writable=False, readable=False),
)

db.define_table("relation",
    Info(),
    Field("member_id", "reference info", notnull=True),
    Field("role"),
)

# db.define_table("data_source",
#     Field("source_name"),
#     migrate = False
# )



# -*- coding: utf-8 -*-

from mptools.dbutils.wiz import Wiz

from .. import settings

def initdb():
    dbwiz = Wiz(path = settings.DB_FOLDER, cache = None)
    dbwiz.collect(settings.DB_URI, 'postgis', 'fuzzystrmatch', 'pgh3')
    dbwiz.setup()
    return dbwiz



# -*- coding: utf-8 -*-

import os
from pathlib import Path
from py4web.core import required_folder
from . import __name__ as PACKAGENAME

HOME = str(Path.home())

# db settings
APP_FOLDER = os.path.join(HOME, PACKAGENAME)

# DB_FOLDER:    Sets the place where migration files will be created
#               and is the store location for SQLite databases
DB_FOLDER = required_folder(APP_FOLDER, "databases")
DB_URI = None
DB_POOL_SIZE = 10
DB_MIGRATE = True
# DB_FAKE_MIGRATE = False  # maybe?

# location where to store uploaded files:
# UPLOAD_FOLDER = os.path.join(APP_FOLDER, "uploads")

# logger settings
LOGGERS = [
    "info:stdout"
]  # syntax "severity:filename" filename can be stderr or stdout

MATERIALIZED_VIEWS = []

PATHROOT = 'planet'

IS_H3_INSTALLED = False



# -*- coding: utf-8 -*-

from os import path
import inspect

def setup():
    here = path.dirname(path.abspath(inspect.getfile(inspect.currentframe()))) # script directory
    with open(path.join(here, 'sql/mercantile.sql')) as pgmercantile:
        sql = pgmercantile.read()
    from ..common import db
    db.executesql(sql)
    db.commit()



# -*- coding: utf-8 -*-

SOURCES = """SELECT DISTINCT ON (info.source_name)
    info.id,
    info.source_name as name
FROM info
ORDER BY info.source_name ASC, info.id ASC
"""

ADDRESSES = """
SELECT DISTINCT ON (info.tags->>'addr:street')
    info.id as id,
    info.source_name as source_name,
    info.tags->>'addr:city' as city,
    info.tags->>'addr:street' as street
FROM info
WHERE
    info.tags::jsonb ? 'addr:street'
ORDER BY
    info.tags->>'addr:street'
"""

HOUSENUMBERS = """
SELECT
    info.id as id,
    info.source_name as source_name,
    info.tags->>'addr:housenumber' as housenumber,
    info.tags->>'addr:street' as street,
    info.tags->>'addr:city' as city
FROM info
WHERE
    info.tags::jsonb ? 'addr:street'
"""

POINTS = """
SELECT
    info.id AS id,
    info.source_name as source_name,
    sources.id as src_id,
    node.geom as geom,
    info.source_id as source_id,
    info.tags as tags,
    info.properties as properties,
    json_build_array(ST_X(node.geom),ST_Y(node.geom)) as crds,
    node.id as node_id
FROM
    info,
    node,
    sources
WHERE
    sources.name=info.source_name AND
    node.info_id=info.id AND
    ((info.tags IS NOT NULL) OR (info.properties IS NOT NULL))
"""

WAYS = """
SELECT
    info.id AS id,
    ST_MakeLine(node.geom ORDER BY way_node.sorting) as geom,
    ST_Centroid(ST_MakeLine(node.geom ORDER BY way_node.sorting)) as centroid,
    ST_IsClosed(ST_MakeLine(node.geom ORDER BY way_node.sorting)) as is_closed,
    count(node.id) as length,
    info.source_name,
    sources.id as src_id,
    info.source_id as source_id,
    info.tags as tags,
    info.properties as properties
FROM
    info,
    node,
    way_node,
    sources
WHERE
    sources.name=info.source_name AND
    way_node.info_id=info.id AND
    way_node.node_id=node.id
GROUP BY
    info.id, sources.id
"""

GRAPH = """
SELECT
    -- start_node.id as id,
    format('%s-%s', start_node.id, end_node.id) as id,
    way_info.source_name,
    sources.id as src_id,
    way_info.source_id,
    ST_MakeLine(ARRAY[start_node.geom, end_node.geom]) as geom,
    start_info.id as sinfo_id,
    start_node.id as snode_id,
    start_info.tags as stags,
    end_info.id as tinfo_id,
    end_node.id as tnode_id,
    end_info.tags as ttags,
    way_info.tags as tags,
    way_info.properties as properties,
    ST_Distance(ST_Transform(start_node.geom, 3857), ST_Transform(end_node.geom, 3857)) as len,
    start_node.geom as snode,
    end_node as tnode
FROM
    info as way_info,
    info as start_info,
    info as end_info,
    node as start_node,
    node as end_node,
    way_node as start_way,
    way_node as end_way,
    sources
WHERE
    sources.name=way_info.source_name AND
    way_info.id = start_way.info_id AND
    start_info.id = start_node.info_id AND
    end_info.id = end_node.info_id AND
    start_way.node_id = start_node.id AND
    end_way.node_id = end_node.id AND
    (way_info.tags::jsonb?'highway')::boolean AND
    (start_way.sorting+1 = end_way.sorting)
"""

SPLITTEDSEGMENTSNODES = """
SELECT
    nextval('snodesseq') as id,
    -- insta_id(node.info_id) as id,
    node.info_id as node_id,
    wayinfo.id as way_id,
    relinfo.id as rel_id,
    count(node.id) OVER (PARTITION BY way_node.info_id) as length,
    ST_MakeLine(node.geom) OVER (PARTITION BY way_node.info_id ORDER BY way_node.sorting) as geom,
    node.geom as node,
    way_node.sorting as sorting,
    relinfo.source_name AS source_name,
    sources.id as src_id,
    relinfo.tags AS tags,
    relinfo.properties AS properties,
    relinfo.source_id AS source_id,
    relation.role as role,
    array_agg(relation.role) OVER (PARTITION BY relation.info_id) as all_roles
FROM
    info as relinfo,
    info as wayinfo,
    node,
    way_node,
    relation,
    sources
WHERE
    sources.name=relinfo.source_name AND
    node.id = way_node.node_id AND
    relation.info_id = relinfo.id AND
    way_node.info_id=wayinfo.id AND
    way_node.info_id = relation.member_id AND
    relinfo.tags::jsonb ?| array['landuse', 'boundary']
"""

SIMPLEPOLYGONS = """
SELECT
    -- format('%s', id),
    nextval('spolyseq') as id,
	ST_MakePolygon(geom) as geom,
    centroid,
    trunc(ST_Area(ST_Transform(ST_MakePolygon(geom), 3857))::numeric, 2) as area,
	source_name,
    src_id,
	source_id,
	tags::jsonb,
	properties::jsonb
FROM (
    SELECT * FROM ways
    WHERE length > 2 AND is_closed AND
        ((tags::jsonb?'area' AND tags->>'area'<>'no') OR tags::jsonb ?| array['landuse', 'boundary', 'building', 'leisure', 'natural'])
    ) as filtered_way
"""

REBUILDEDPOLYGONPATHS = """
SELECT
    -- format('%s-%s', segment_points.rel_id, nextval('mypolyseq')) as id,
    -- format('%s-%s', segment_points.rel_id::text, md5(random()::text || clock_timestamp()::text)::uuid) as id,
    nextval('rpathseq') as id,
    -- min(segment_points.way_id) as way_id,
    -- insta_id(segment_points.rel_id) as id,

    (ST_Dump(ST_LineMerge(ST_Union(segment_points.geom order by segment_points.rel_id)))).geom as geom,
    ST_IsClosed((ST_Dump(ST_LineMerge(ST_Union(segment_points.geom order by segment_points.rel_id)))).geom) as is_closed,
    sum(segment_points.length) as length,
    segment_points.source_name,
    segment_points.src_id,
    segment_points.tags::jsonb,
    segment_points.properties::jsonb,
    segment_points.source_id
FROM segment_points
WHERE 'outer' = ANY(all_roles)
GROUP BY
    segment_points.rel_id, segment_points.source_name, segment_points.src_id, segment_points.source_id,
    segment_points.tags::jsonb, segment_points.properties::jsonb
"""

REBUILDEDPOLYGONS = """
SELECT
    nextval('spolyseq') as id,
    -- way_id,
    ST_MakePolygon(geom) as geom,
    ST_Centroid(geom) as centroid,
    trunc(ST_Area(ST_Transform(ST_MakePolygon(geom), 3857))::numeric, 2) as area,
    source_name,
    src_id,
    source_id,
    tags,
    properties
FROM (SELECT * FROM {rpath} WHERE is_closed) as closed_paths
""".format

POLYS = "(SELECT * FROM {spolys}) UNION (SELECT * FROM rpolys);".format



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import mercantile as mc
from math import modf
import pyproj
import geojson

BASE_DIM = 155


class Bbox(object):
    """docstring for Bbox."""

    def __init__(self, minx, miny, maxx, maxy):
        """
        minx @float : Min longitude
        miny @float : Min latitude
        maxx @float : Max longitude
        maxy @float : Max latitude
        """
        super(Bbox, self).__init__()
        self.minx = minx
        self.miny = miny
        self.maxx = maxx
        self.maxy = maxy

    def dims(self, crs="epsg:3857"):
        osmproj = pyproj.Proj(crs)
        minx, miny = osmproj(self.minx, self.miny)
        maxx, maxy = osmproj(self.maxx, self.maxy)
        # transformer = Transformer.from_crs(self.crs, crs)
        # minx, miny = transformer.transform(self.minx, self.miny)
        # maxx, maxy = transformer.transform(self.maxx, self.maxy)
        return maxx-minx, maxy-miny

    @property
    def osm(self):
        keys = ('w', 's', 'e', 'n')
        return dict(zip(keys, map(str, (self.minx, self.miny, self.maxx, self.maxy,)))) # minx, miny, maxx, maxy

    @property
    def as_gj_polygon(self):
        return geojson.Polygon([[
            (self.minx,self.maxy,),
            (self.maxx,self.maxy,),
            (self.maxx,self.miny,),
            (self.minx,self.miny,),
            (self.minx,self.maxy,),
        ]])

    def __repr__(self):
        return str(self.osm)


def get_base_tile(lon, lat, bdim=BASE_DIM):
    """
    lon  @float : Center longitude.
    lat  @float : Center latitude.
    bdim @float : Base time dimension in meters.

    returns:
        tt @mercantile.tile : Tile nearest to desidered dimension coontaining the
                              point of given coordinates.
        zoom_level @integer : The tile zoom level.
    """
    max_zoom_level = 30
    max_area = bdim**2
    for zoom_level in range(max_zoom_level):
        tt = mc.tile(lon, lat, zoom_level)
        bb = mc.xy_bounds(tt)
        if (bb.right-bb.left)*(bb.top-bb.bottom) < max_area:
            break
    return tt, zoom_level

def __get_box_dim(bt, dist):
    """ """

    ss = mc.xy_bounds(bt)
    b, h = ss.right-ss.left, ss.top-ss.bottom
    real_dim = min([b,h])

    fp, ip = modf(dist/real_dim)
    nn = ip + 1 if fp > 0 else ip
    return int(nn)

def tilebbox(dist, lon, lat, bdim=BASE_DIM, buffer=4, format=None):
    """
    dist     @float : Search distance.
    lon      @float : Center longitude.
    lat      @float : Center latitude.
    bdim     @float : Base tile dimention in meters.
    buffer @integer : Buffer dimention in times of bdim.
    format  @string : Output format

    returns:
        upper left corner coordinates and bottom right cooodinates in terms of
        * [[minlon, maxlat], [maxlon, minlat]]
        * {'w': minlon, 's': minlat, 'e': maxlon, 'n': maxlat} in case OSM format is required
    """

    bt, _ = get_base_tile(lon, lat, bdim)

    nn = __get_box_dim(bt, dist)

    ul = mc.ul(bt.x-(nn+buffer), bt.y-(nn+buffer), bt.z)
    br = mc.ul(bt.x+(nn+buffer)+1, bt.y+(nn+buffer)+1, bt.z)

    if format is None:
        return Bbox(ul.lng, br.lat, br.lng, ul.lat)
    elif format == 'osm':
        keys = ('w', 's', 'e', 'n')
        return dict(zip(keys, map(str, (ul.lng, br.lat, br.lng, ul.lat,)))) # minx, miny, maxx, maxy
    else:
        raise NotImplementedError

def boxtiles(dist, lon, lat, bdim=BASE_DIM, buffer=4):
    """ """
    bt, _ = get_base_tile(lon, lat, bdim)
    nn = __get_box_dim(bt, dist)//2

    for xx in range(-(nn+buffer-1), (nn+buffer+1)):
        for yy in range(-(nn+buffer-1), (nn+buffer+1)):
            yield bt.x+xx, bt.y+yy, bt.z,
    #         bounds = mc.bounds(bt.x+xx, bt.y+yy, bt.z)
    #         out[(bt.x+xx, bt.y+yy, bt.z,)] = Bbox(
    #             bounds.west,
    #             bounds.south,
    #             bounds.east,
    #             bounds.north
    #         )
    # return out



# -*- coding: utf-8 -*-

import os
import hashlib
import geojson
import overpy
# from lxml import etree
from xml.etree import ElementTree as etree
from itertools import groupby
from time import sleep
from datetime import datetime
from functools import reduce
from urllib.request import urlopen
from urllib.error import HTTPError
from io import BytesIO

from .geom import Bbox, BASE_AREA_DIMENSION
from ...tools.tile import BASE_DIM, tilebbox

PY2 = False

class MaxRetriesReached(overpy.exception.OverPyException):
    """ Courtesy of: https://github.com/DinoTools/python-overpy/blob/31c6e689c0d49d6617e020597914f47a0bc98f04/overpy/exception.py#L40
    Raised if max retries reached and the Overpass server didn't respond with a result.
    """
    def __init__(self, retry_count, exceptions):
        self.exceptions = exceptions
        self.retry_count = retry_count

    def __str__(self):
        return "Unable get any result from the Overpass API server after %d retries." % self.retry_count


overpy.exception.MaxRetriesReached = MaxRetriesReached
overpy.Overpass.default_max_retry_count = 3
overpy.Overpass.default_retry_timeout = 10
# overpy.Overpass.max_retry_count = 5
overpy.Overpass.retry_timeout = 10

class Turbo(object):
    """ OSM Overpass Turbo utils """

    default_base_resource_path = ""

    def __init__(self, *args, **kw):
        self.__cache__ = {}
        self.api = overpy.Overpass(*args, **kw)

    def __raw_call__(self, query):
        """ Reimplements the call to Overpass Turbo web service for DEBUG purposes only.
        Courtesy of: https://github.com/DinoTools/python-overpy/blob/4b5ace5baf854dd84dbfea955d0c67f602bd754d/overpy/__init__.py#L113
        Query the Overpass API
        :param String|Bytes query: The query string in Overpass QL
        :return: The parsed result
        :rtype: overpy.Result
        """

        try:
            max_retry_count = self.api.max_retry_count
        except AttributeError:
            max_retry_count = self.api.default_max_retry_count

        try:
            retry_timeout = self.api.retry_timeout
        except AttributeError:
            retry_timeout = self.api.default_retry_timeout

        read_chunk_size = 4096

        if not isinstance(query, bytes):
            query = query.encode("utf-8")

        retry_num = 0
        retry_exceptions = []
        do_retry = True if max_retry_count > 0 else False
        while retry_num <= max_retry_count:
            if retry_num > 0:
                sleep(retry_timeout)
            retry_num += 1
            t0 = datetime.now()
            try:
                f = urlopen(self.api.url, query)
            except HTTPError as e:
                f = e
            else:
                dt = datetime.now()-t0

            response = f.read(read_chunk_size)
            while True:
                data = f.read(read_chunk_size)
                if len(data) == 0:
                    break
                response = response + data
            f.close()

            if f.code == 200:

                if PY2:
                    http_info = f.info()
                    content_type = http_info.getheader("content-type")
                else:
                    content_type = f.getheader("Content-Type")

                if content_type in ("application/json", "application/osm3s+xml",):
                    return response, dt

                e = overpy.exception.OverpassUnknownContentType(content_type)
                if not do_retry:
                    raise e
                retry_exceptions.append(e)
                continue

            if f.code == 400:
                e = overpy.exception.OverpassBadRequest(query, msgs=None)
                if not do_retry:
                    raise e
                retry_exceptions.append(e)
                continue

            if f.code == 429:
                e = overpy.exception.OverpassTooManyRequests
                if not do_retry:
                    raise e
                retry_exceptions.append(e)
                continue

            if f.code == 504:
                e = overpy.exception.OverpassGatewayTimeout
                if not do_retry:
                    raise e
                retry_exceptions.append(e)
                continue

        raise overpy.exception.MaxRetriesReached(retry_count=retry_num, exceptions=retry_exceptions)

    raw_call = __raw_call__

    def __fake_call__(self, query, *args, **kw):
        """ Caches responses for DEBUG purposes only """

        queryhash = hashlib.new('sha224')
        queryhash.update(query)
        querychecksum = queryhash.hexdigest()

        rp = os.path.join(self.default_base_resource_path, "resources", querychecksum)

        resource_file = os.path.normpath(rp)

        if not os.path.isfile(resource_file):
            if not os.path.exists(os.path.dirname(resource_file)):
                try:
                    os.makedirs(os.path.dirname(resource_file))
                except OSError as exc: # Guard against race condition
                    if exc.errno != errno.EEXIST:
                        raise

            raw_data, dt = self.__raw_call__(query)

            if not raw_data is None:
                with open(resource_file, "w") as res:
                    res.write(raw_data)
            data_dim = os.stat(resource_file).st_size
            delta_sec = dt.total_seconds()
            logger.info("Retrieved {} bytes in {} seconds (at a speed of {} Bps)".format(
                data_dim, delta_sec,
                data_dim/delta_sec
            ))

        assert os.path.isfile(resource_file), "File Not Found!"
        with open(resource_file, mode='rb') as myxml:
            body = myxml.read()
        return self.api.parse_json(body) #.nodes

    def __call__(self, query, *args, **kw):
        """
        query @string : The OSM query ready to be submitted to the Overpass web service.
        """

        queryhash = hashlib.new('sha224')
        queryhash.update(query)
        querychecksum = queryhash.hexdigest()

        try:
            response = self.__cache__[querychecksum]["response"]
            assert not response is None, "This should never happen, why it happens?"
        except (KeyError, AssertionError,):
            for t in range(1, 5):
                try:
                    response = self.api.query(query, *args, **kw) #.nodes
                    assert not response is None, "This should never happen, why it happens?"
                except (
                    overpy.exception.OverpassTooManyRequests,
                    overpy.exception.OverpassGatewayTimeout,
                    AssertionError,
                ) as err:
                    if t==5:
                        raise
                    else:
                        sleep(t*overpy.Overpass.retry_timeout)
                        continue
                else:
                    self.__cache__[querychecksum] = {
                        "response": response,
                        "query": query
                    }
                    return self.__cache__[querychecksum]["response"]

        else:
            return response # self.__cache__[querychecksum]["response"]

    def nodes(self, *args, **kw):
        return self(*args, **kw)

    @staticmethod
    def build_query(qconditions, gtypes=None, timeout=60, pretty_print=False, maxsize=None):
        """ Returns the Overpass turbo query in XML format
        qconditions  @list : Query conditions;
        gtypes       @list : One or more values between 'node', 'way', 'relation';
        timeout   @int/str : The query timeout;
        pretty_print @bool : If true returns the indented XML output for debug purposes.

        Examples:

        qconditions = [{
            "query": [[{"k": ..., "modv": ..., "v/regv": ...}, ...], ...],
            "bbox": ...,
            "gtypes": [...], # Optional. Possible values: "node", "way", "relation"
            # "amplitude": 0,
            "newer": "%Y-%m-%ddT%H:%M:%SZ" #
        }]
        """

        default_gtype_values = ("node", "way", "relation",)

        def __check__(geom):
            msg_template = "{geom} is not a valid OSM entity type (possible values are: {geoms})."
            gtypes
            assert (geom in default_gtype_values), msg_template.format(
                geom = geom,
                geoms = ','.join(default_gtype_values)
            )

        if gtypes is None:
            gtypes = ("node",)
        else:
            map(__check__, gtypes)

        Root = etree.Element(
            "osm-script",
            output = "json",
            timeout = "{0:d}".format(timeout),
            **{"element-limit": maxsize for x in 'a' if not maxsize is None}
        )
        Union = etree.SubElement(Root, "union", into="_")

        _recurse = {
            "way": ["down"],
            "relation": ["down"]
        }

        def _append(gtype, query, bbox=None, newer_than=None):
            for _union in query:
                Query = etree.SubElement(Union, "query", into="_", type=gtype)
                for _intersection in _union:
                    etree.SubElement(Query, "has-kv", **_intersection)
                if not newer_than is None:
                    etree.SubElement(Query, "newer", than=newer_than)
                for recurse_type in _recurse.get(gtype, []):
                    etree.SubElement(Union, "recurse", into="_", type=recurse_type, **{"from": "_"})
                if not bbox is None:
                    etree.SubElement(Query, "bbox-query", **bbox)

        for cond in qconditions():
            try:
                _gtypes = cond.pop('gtypes')
            except KeyError:
                _gtypes = gtypes

            for _type in _gtypes:
                _append(_type, **cond)

        etree.SubElement(Root, "print",
            e="",geometry="skeleton", limit="", mode="meta", n="", order="id", s="", w="",
            **{"from": "_"}
        )
        ff = BytesIO()
        et = etree.ElementTree(Root)
        et.write(ff, encoding='utf-8', xml_declaration=True)
        return ff.getvalue().decode()
        # return b'<?xml version="1.0" encoding="UTF-8"?>'+etree.tostring(Root, pretty_print=pretty_print)

    @staticmethod
    def optimize_centralized_query_by_base_tile(lon, lat, qconditions, bdim=BASE_DIM, buffer=3, newer_than=None):
        """
        lon @float :
        lat @float :
        qconditions  @list : Query conditions given by distance from point;
        bdim @float : Base tile dimention;
        buffer @integer : Number of buffer tiles;
        newer_than @string : "%Y-%m-%dT%H:%M:%SZ"

        Examples:

        qconditions = [{
            "query": [[{"k": ..., "modv": ..., "v/regv": ...}, ...], ...],
            "distance": ...,
            "gtypes": [...], # Optional. Possible values: "node", "way", "relation", "way-node", node-relation", "relation-way", "relation-relation", "relation-backwards"
            ""
        }, ...]
        """

        grouping_func = lambda cnd: (int(cnd['distance']//bdim), cnd['gtypes'],)

        def _main():
            _grouped_queryes_ = groupby(sorted(qconditions, key=grouping_func), key=grouping_func)
            for gd, _cnds in _grouped_queryes_:
                _dist, gtypes = gd
                max_dist = (_dist+1)*bdim
                bbox = tilebbox(max_dist, lon, lat, bdim=bdim, buffer=buffer, format='osm')
                allcnds = reduce(lambda a,b: a+b, (tuple(i["query"]) for i in _cnds))
                yield {
                    "bbox": bbox,
                    "query": tuple(dict([(tuple(map(lambda j: tuple(j.items()), i)), i,) for i in allcnds]).values()),
                    "gtypes": gtypes,
                    "newer_than": newer_than
                }

        return _main

    @staticmethod
    def optimize_centralized_query(lon, lat, qconditions, buffer=None):
        """
        qconditions  @list : Query conditions given by distance from point;

        Examples:

        qconditions = [{
            "query": [[{"k": ..., "modv": ..., "v/regv": ...}, ...], ...],
            "distance": ...
        }, ...]
        """

        def _main():

            _grouped_queryes_ = groupby(
                qconditions,
                key = lambda cnd: float("{distance:d}".format(**cnd))
            )

            for gd, _cnds in _grouped_queryes_:
                cnds = (tuple(i["query"]) for i in _cnds)
                allcnds = sum(cnds, ())
                yield {
                    "bbox": Bbox.bufferfactory(
                        lon=lon, lat=lat, crss=("epsg:4326",),
                        buffer = gd + (buffer if not buffer is None else BASE_AREA_DIMENSION)
                    ).osmbbox,
                    "query": tuple(dict([(tuple(map(lambda j: tuple(j.items()), i)), i,) for i in allcnds]).values())
                }

        return _main

    def iter(self, lon, lat, qconditions, buffer=None, gtypes=None, **kw):
        """ """
        query = self.build_query(
            self.optimize_centralized_query(lon, lat,
                qconditions,
                buffer=buffer
            ),
            gtypes=gtypes
        )
        return self(query, **kw)

    @staticmethod
    def check(tags, k, modv=None, **kw):
        """
        Evalluates whether the OSM node properties respect the single filter condition.
        Returns boolean.

        properties @dict : The node tags (at leas object must support __getitem__);
        k        @string : The OSM Overpass Turbo condition tag key;
        modv     @string : The OSM Overpass Turbo condition operator (None/"not").
                           If modv is None a value for "regv" must be provided
                           (NOTE: at the moment the only supported value for
                           "regv" is the wildcard ".")
        v        @string : (If provided) is the OSM Overpass Turbo queried tag value.

        || WARNINNG! It's a very raw implementation but enough good for       ||
        || our actual purposes.                                               ||
        """

        def _check():

            try:
                rvalue = tags[k]
            except KeyError:
                return False
            else:
                try:
                    cvalue = kw["v"]
                except KeyError:
                    try:
                        cvalue = kw["regv"]
                    except KeyError:
                        return True
                    else:
                        if cvalue==".":
                            return True
                        else:
                            raise NotImplementedError("Filter not yet supported.")
                else:
                    return not cvalue or cvalue==rvalue

        if not modv:
            return _check()
        elif modv=="not":
            return not _check()
        else:
            raise NotImplementedError("Operator not yet supported")



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py4geo",
    version="1.0.7",
    author="Manuele Pesenti",
    author_email="manuele@inventati.org",
    description="My geo package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/manuelep/py4geo",
    project_urls={
        "Bug Tracker": "https://github.com/manuelep/py4geo/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    package_data={"": ['*.sql', '*.md']},
    python_requires=">=3.6",
    install_requires=[
        "overpy",
        "shapely",
        "ujson",
        "pyproj",
        "mercantile",
        "lxml",
        "geojson",
        "hashids",
        "tqdm",
        "geomet",
        "h3",
        "mptools>=1.0.7",
        "py4vtile",
        "supermercado",
        "osm2geojson"
    ]
)



# -*- coding: utf-8 -*-

from pydal import geoPoint
from .pgcopy import BulkCopyer
from .base import WTF, BaseParser, BaseCopier
from hashids import Hashids
myhashids = Hashids()
from tqdm import tqdm

class __CommonMethods__(object):
    """docstring for __FeatureParser__."""

    def parse_feature(self, feature, merge=False):
        """ """
        method_name = feature["geometry"]["type"]
        try:
            method = getattr(self, "_parse{}".format(method_name))
        except (AttributeError, NotImplementedError,) as err:
            raise NotImplementedError(feature["geometry"]["type"])
        else:
            info_id = method(feature)
            if info_id is None:
                # import pdb; pdb.set_trace()
                raise WTF()

            return info_id


class NodeParser(BaseParser, __CommonMethods__):
    """docstring for GeojsonParser."""

    def __init__(self, db, source_name, tags_on_update="replace", properties_on_update="replace"):
        """
        db @DAL : The database in wich info, node, way_node and relation tables are defined.
        source_name @string : Source name
        tags_on_update @string : Supported values:
            - "replace": Replace tags
            - "update": Update tags
        properties_on_update @string : See tags_on_update.
        """
        super(NodeParser, self).__init__(db, source_name)
        self.tags_on_update = tags_on_update
        self.properties_on_update = properties_on_update

    def _parsePoint(self, feature, merge=False):
        """ """
        info_id = self._save_info(feature["id"],
            gtype = 'node',
            tags = feature.get('tags'),
            properties = feature["properties"],
            merge = merge
        )
        coordinates = feature["geometry"]["coordinates"]
        node_id = self._save_node(info_id, *coordinates)
        return info_id

    def parse_features(self, features, merge=False):
        """
        features api:
            Standard geojson feature with additional optional property "tags"
            i.e.: {
              "type": "Feature",
              "geometry": {
                "type": "Point",
                "coordinates": [125.6, 10.1]
              },
              "properties": {
                "name": "Dinagat Islands"
              },
              "tags": {"<tag name>": <tag value>} # OPTIONAL NON GEOJSON STANDARD
            }
        """
        for feature in tqdm(features, desc="Parsing features"):
            info_id = self.parse_feature(feature, merge=merge)

    parse = parse_features


class WayParser(NodeParser):

    def _save_way(self, info_id, feat_id, coordinates, wn=0):
        for sorting, xy in enumerate(coordinates):
            nsid = '{}-{}'.format(feat_id, myhashids.encode(wn, sorting))
            node_info_id = self._save_info(nsid, gtype='node')
            node_id = self._save_node(node_info_id, *xy)
            data = dict(info_id=info_id, node_id=node_id, sorting=sorting)
            self._save_way_node(**data)

    def _parseWay(self, feature, merge=False):

        info_id = self._save_info(
            feature["id"],
            tags = feature.get('tags'),
            properties = feature['properties'],
            gtype="way",
            merge = merge
        )

        if feature['geometry']['type']=='LineString':
            self._save_way(info_id, feature["id"], feature["geometry"]['coordinates'])
        elif feature['geometry']['type']=='Polygon':
            self._save_way(info_id, feature["id"], feature["geometry"]['coordinates'][0])
            crdlen = len(feature["geometry"]['coordinates'])

        return info_id

    _parseLineString = _parseWay


class PolygonParser(WayParser):

    def _parseMultiPolygon(self, feature, merge=False):

        info_id = self._save_info(
            feature["id"],
            tags = dict(feature.get('tags', {}), type='multipolygon'),
            properties = feature['properties'],
            gtype = "relation",
            merge = merge
        )

        for wn,waynodes in enumerate(feature["geometry"]['coordinates']):
            swid = '{}-{}'.format(feature["id"], myhashids.encode(wn))
            way_info_id = self._save_info(swid, gtype='way')
            self._save_way(way_info_id, swid, waynodes, wn=wn)

            self._save_relation(
                info_id, way_info_id,
                role = 'outer' if wn==0 else 'inner'
            )

        return info_id

    def _parsePolygon(self, feature, merge=False):
        """ """

        check = len(feature["geometry"]['coordinates'])
        if check == 1:
            gtype = 'way'
        elif check > 1:
            gtype = 'relation'
        else:
            raise ValueError

        if gtype=='way':
            info_id = self._parseWay(feature, merge=merge)
        # gtype == 'relation'
        else:
            info_id = self._parseMultiPolygon(feature, merge=merge)

        return info_id


class NodeCopier(BaseCopier, __CommonMethods__):
    """docstring for NodeCopier."""

    def _parsePoint(self, feature, **__):
        """ """
        info_id = self._save_info(feature["id"],
            gtype = 'node',
            tags = feature.get('tags'),
            properties = feature["properties"]
        )
        node_id = self._save_node(info_id, *feature["geometry"]["coordinates"])
        return info_id

    def parse(self, features, **__):
        """
        features api:
            Standard geojson feature with additional optional property "tags"
            i.e.: {
              "type": "Feature",
              "geometry": {
                "type": "Point",
                "coordinates": [125.6, 10.1]
              },
              "properties": {
                "name": "Dinagat Islands"
              },
              "tags": {"<tag name>": <tag value>} # OPTIONAL NON STANDARD OSM LIKE TAGS
            }
        """

        with BulkCopyer(self.db.node) as self._insnodes, \
            BulkCopyer(self.db.info) as self._insinfo:

            for feature in features:
                info_id = self.parse_feature(feature)


class WayCopier(NodeCopier, WayParser):
    """ """

    def parse(self, features, **__):
        """
        features api:
            Standard geojson feature with additional optional property "tags"
            i.e.: {
              "type": "Feature",
              "geometry": {
                "type": "Point",
                "coordinates": [125.6, 10.1]
              },
              "properties": {
                "name": "Dinagat Islands"
              },
              "tags": {"<tag name>": <tag value>} # OPTIONAL NON STANDARD OSM LIKE TAGS
            }
        """

        with BulkCopyer(self.db.way_node) as self._insways, \
            BulkCopyer(self.db.node) as self._insnodes, \
            BulkCopyer(self.db.info) as self._insinfo:

            for feature in features:
                info_id = self.parse_feature(feature)


class PolygonCopier(WayCopier, PolygonParser):
    """docstring for PolygonCopier."""

    # def _parseMultiPolygon(self, feature):
    #
    #
    # def _parsePolygon(self, feature):
    #     """ """
    #
    #     check = len(feature["geometry"]['coordinates'])
    #     if check == 1:
    #         gtype = 'way'
    #     elif check > 1:
    #         gtype = 'relation'
    #     else:
    #         raise ValueError
    #
    #     # info_id = self.save_info(
    #     #     feature["id"],
    #     #     feature["properties"],
    #     #     gtype = gtype
    #     # )
    #
    #     _info_ = lambda **kw: dict(
    #         sid = feature["id"],
    #         gtype = gtype,
    #         tags = dict(feature.get('tags', {}), **kw) or None,
    #         properties = feature['properties']
    #     )
    #
    #     if gtype=='way':
    #         info_id = self._save_info(**_info_())
    #         for sorting, xy in enumerate(feature["geometry"]['coordinates'][0]):
    #
    #             nsid = '{}-{}'.format(feature["id"], myhashids.encode(0, sorting))
    #
    #             node_info_id = self._save_info(nsid, gtype='node')
    #             node_id = self._save_node(node_info_id, *xy)
    #             data = dict(
    #                 info_id = info_id,
    #                 node_id =  node_id,
    #                 sorting = sorting
    #             )
    #             self._insways.writerow(data)
    #     # gtype == 'relation'
    #     else:
    #         try:
    #             info_id = self._save_info(**_info_(type='multipolygon'))
    #         except Exception as err:
    #             import pdb; pdb.set_trace()
    #         for wn,waynodes in enumerate(feature["geometry"]['coordinates']):
    #
    #             wsid = '{}-{}'.format(feature["id"], myhashids.encode(wn))
    #             way_info_id = self._save_info(wsid, gtype='way')
    #
    #             if wn == 0:
    #                 data = dict(info_id=info_id, member_id=way_info_id, role='outer')
    #                 self._insrelation.writerow(dict(
    #                     suid = self.db.relation.suid.compute(data),
    #                     **data
    #                 ))
    #                 # self._insrelation.writerow(dict(info_id=info_id, way_id=way_info_id, outer_way_id="NULL"))
    #                 # _outer_way_id=int(way_info_id)
    #             else:
    #                 data = dict(info_id=info_id, member_id=way_info_id, role='inner')
    #                 self._insrelation.writerow(dict(
    #                     suid = self.db.relation.suid.compute(data),
    #                     **data
    #                 ))
    #                 # self._insrelation.writerow(dict(info_id=info_id, way_id=way_info_id, outer_way_id=_outer_way_id))
    #
    #             for sorting, xy in enumerate(waynodes):
    #                 nsid = '{}-{}'.format(feature["id"], myhashids.encode(wn, sorting))
    #                 assert nsid
    #                 node_info_id = self._save_info(nsid, gtype='node')
    #                 node_id = self._save_node(node_info_id, *xy)
    #                 data = dict(
    #                     info_id = way_info_id,
    #                     node_id =  node_id,
    #                     sorting = sorting
    #                 )
    #                 self._insways.writerow(data)
    #
    #     return info_id

    # def parse_(self, features, **__):
    #     """
    #     features api:
    #         Standard geojson feature with additional optional property "tags"
    #         i.e.: {
    #           "type": "Feature",
    #           "geometry": {
    #             "type": "Point",
    #             "coordinates": [125.6, 10.1]
    #           },
    #           # OPTIONAL NON STANDARD OSM LIKE TAGS
    #           "properties": {"name": "Dinagat Islands"},
    #           # STANDARD OSM LIKE TAGS
    #           "tags": {"<tag name>": <tag value>}
    #         }
    #     """
    #
    #     with BulkCopyer(self.db.relation) as self._insrelation, \
    #         BulkCopyer(self.db.way_node) as self._insways, \
    #         BulkCopyer(self.db.node) as self._insnodes, \
    #         BulkCopyer(self.db.info) as self._insinfo:
    #
    #         for feature in features:
    #             yield self.parse_feature(feature)

    def parse(self, features, **__):
        """
        features api:
            Standard geojson feature with additional optional property "tags"
            i.e.: {
              "type": "Feature",
              "geometry": {
                "type": "Point",
                "coordinates": [125.6, 10.1]
              },
              # OPTIONAL NON STANDARD OSM LIKE TAGS
              "properties": {"name": "Dinagat Islands"},
              # STANDARD OSM LIKE TAGS
              "tags": {"<tag name>": <tag value>}
            }
        """

        with BulkCopyer(self.db.relation) as self._insrelation, \
            BulkCopyer(self.db.way_node) as self._insways, \
            BulkCopyer(self.db.node) as self._insnodes, \
            BulkCopyer(self.db.info) as self._insinfo:

            for feature in tqdm(features):
                info_id = self.parse_feature(feature)



# -*- coding: utf-8 -*-

from ..common import logger
from .optutils.base import Turbo
from . import io
import json
import osm2geojson
from ..tools.tile import Bbox
# from .tile import boxtiles

from tqdm import tqdm
from sys import stdout, getsizeof #, _getframe
from mptools.dataformat import smartbytes
import geojson
from supermercado.burntiles import burn
import mercantile

T = lambda s: s

get_uri = lambda xtile, ytile, zoom: "{xtile:d}/{ytile:d}/{zoom:d}".format(xtile=xtile, ytile=ytile, zoom=zoom)

class barFormat(object):
    """docstring for barFormat."""

    base_fmt = 'desc', 'n', 'rate', 'rate_inv', 'elapsed', 'remaining', 'unit',
    augm_fmt = 'percentage', 'total',

    @staticmethod
    def cast(value):
        """ """
        try:
            return json.loads(value)
        except ValueError:
            return value

    @classmethod
    def template(cls, augm=False):
        fmt = lambda args: 'progress '+'{{{}}}'.format('};{'.join(args))
        return fmt(cls.headers(augm=augm))

    @classmethod
    def headers(cls, augm=True):
        if augm:
            return cls.base_fmt+cls.augm_fmt
        else:
            return cls.base_fmt

    @classmethod
    def parse(cls, *rows):
        """ """
        return filter(lambda shit: shit, map(
            lambda row: dict(zip(cls.headers(), map(cls.cast, filter(lambda v: v.strip(), row.split(';'))))),
            rows
        ))

class MyTqdm(dict):
    """docstring for MyTqdm."""

    def __init__(self, desc, **kwargs):
        super(MyTqdm, self).__init__(**kwargs)
        tot = sum(map(len, kwargs.values()))
        self.tqdm = tqdm(
            desc = desc,
            total = tot,
            # bar_format = barFormat.template(not not tot),
            # file = stdout
        )

    def set_description(self, *args, **kwargs):
        return self.tqdm.set_description(*args, **kwargs)

    def __getattr__(self, key):
        def __main():
            for v in self[key]:
                self.tqdm.update(1)
                yield v
        return [x for x in __main()]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.tqdm.close()

def fetch_and_log_from_osm(query, pgcopy=False):
    """
    query @string : The OSM filter query; Overpass QL or XML.
    """

    # out_task_id = web2py_uuid()
    # plugins.planet.logger.info("### {} >>>".format(out_task_id))
    # plugins.planet.logger.info(query)

    turbo = Turbo()
    data = turbo(query.encode())

    nodes = list(tqdm(
        data.nodes,
        desc = str(T('Downloading nodes from osm')),
        # bar_format = barFormat.template(True if data.nodes else False),
        # file = stdout
    ))

    ways = list(tqdm(
        data.ways,
        desc = str(T('Downloading ways from osm')),
        # bar_format = barFormat.template(True if data.ways else False),
        # file = stdout
    ))

    relations = list(tqdm(
        data.relations,
        desc = str(T('Downloading relations from osm')),
        # bar_format = barFormat.template(True if data.relations else False),
        # file = stdout
    ))

    # dcounter = MyTqdm(
    #     str(T('Saving data to DB')),
    #     nodes = nodes,
    #     ways = ways,
    #     relations = relations
    # )

    # io.osm(dcounter.nodes, dcounter.ways, dcounter.relations, copy=pgcopy)
    io.osm(*map(tqdm, (nodes, ways, relations,)), copy=pgcopy)
    io.db.commit()
    # if request.is_scheduler:
    #     timeLoggerDecorator(plugins.planet.logger)(Put.commit())

    logger.info("Response size: {}".format(smartbytes(getsizeof(data))))
    # plugins.planet.logger.info("<<< {} ###".format(out_task_id[:8]))
    # return out_task_id

class Syncher(object):
    """docstring for Syncher."""
    def __init__(self, xtile, ytile, zoom, uri=None, query=None, gtypes=['node', 'way', 'relation']):
        super(Syncher, self).__init__()
        self.uri = uri or get_uri(xtile, ytile, zoom)
        self.tile = {'x': xtile, 'y': ytile, 'z': zoom}
        tile_bounds = mercantile.bounds(mercantile.quadkey_to_tile(mercantile.quadkey(xtile, ytile, zoom)))
        keys = ('w', 's', 'e', 'n',)
        self.bbox = dict(zip(keys, map(str, (
            tile_bounds.west,
            tile_bounds.south,
            tile_bounds.east,
            tile_bounds.north,
        )))) # minx, miny, maxx, maxy

        if query is None:
            query = [[{"k": "qwertyuiop", "modv": "not", "regv": "."}]]

        self.base_query = {
            'query': query,
            'bbox': self.bbox,
            'gtypes': ['node', 'way', 'relation'],
        }

    def __call__(self, newer_than=None, pgcopy=False):
        """
        newer_than @datetime : Last update timestamp
        """
        if newer_than is None:
            _query = dict(self.base_query)
        else:
            _query = dict(self.base_query,
                newer_than = newer_than.strftime("%Y-%m-%dT%H:%M:%SZ")
            )

        fetch_and_log_from_osm(
            Turbo.build_query(lambda: [_query]),
            pgcopy = pgcopy
        )

def get_polys(featureCollection):
    for feature in featureCollection['features']:
        if feature['geometry']['type'] == 'MultiPolygon':
            for poly in feature['geometry']['coordinates']:
                xfeature = { "type": "Feature", "properties": {}, "geometry": { "type": "Polygon" } }
                xfeature['geometry']['coordinates'] = poly
                yield xfeature
        elif feature['geometry']['type'] == 'Polygon':
            yield feature

def tile2poly(xyz):
    x, y, z = xyz
    bounds = mercantile.bounds(x, y, z)
    bbox = Bbox(bounds.west, bounds.south, bounds.east, bounds.north)
    poly = bbox.as_gj_polygon
    return poly

class OsmShell(object):
    """ Downloads custom data from OSM inside administrative boundary geometry """

    @staticmethod
    def __boundary_filter(*args, **kw):
        def main():
            query = [
                {"k": k, "v": "{}".format(v)} \
            for k,v in kw.items()]+list(args)

            yield {
                # "bbox": Bbox(minlon, minlat, maxlon, maxlat).osm,
                "query": [query],
                "gtypes": ['relation']
            }
        return main

    def __init__(self, name, admin_level=8, zoom=14):
        """
        name         @string : Boiundary name (e.g. Genova, Milano, etc.)
        admin_level @integer : Please refer to https://wiki.openstreetmap.org/wiki/Key:admin_level
            (e.g. 8 Comune or 6 Provincia)
        zoom        @integer : Dimension of tiles used for pixelating the boundary
        """
        super(OsmShell, self).__init__()
        query = Turbo.build_query(self.__boundary_filter(
            {"k": "wikipedia", "regv": "it:"},
            boundary = "administrative",
            admin_level = admin_level,
            name = name
        ))

        turbo = Turbo()
        data_, _ = turbo.__raw_call__(query.encode())
        data = json.loads(data_)
        fc = osm2geojson.json2geojson(data)

        polys = list(get_polys(fc))

        assert len(polys)>0

        self.tiles = burn(polys, zoom)

    @property
    def tileCollection(self):
        """ """
        return geojson.FeatureCollection(list(map(
            lambda gg: geojson.Feature(geometry=gg),
            map(tile2poly, self.tiles)
        )))

    def __call__(self, query=None):
        if query is None:
            # Generic query for downloading everything
            query = [[{"k": "qwertyuiop", "modv": "not", "regv": "."},],]

        for tile in tqdm(self.tiles):
            dbsyncher = Syncher(*tile, query=query)
            dbsyncher()



# -*- coding: utf-8 -*-

from . import settings
from mptools.logger import package_logger

from . import __name__ as PACKAGENAME

logger = package_logger(PACKAGENAME, settings.LOGGERS)

T = lambda s: s

if settings.DB_URI is None:
    db = None
else:
    from py4web import DAL
    # connect to db
    db = DAL(settings.DB_URI,
        folder = settings.DB_FOLDER,
        pool_size = settings.DB_POOL_SIZE,
        migrate = settings.DB_MIGRATE,
        lazy_tables=False, check_reserved=False
    )



# -*- coding: utf-8 -*-

from itertools import zip_longest

# from pydal.helpers.serializers import json as jsondumps
from pydal import geoPoint
from .pgcopy import BulkCopyer
from .base import BaseParser, BaseCopier, WTF
from overpy.exception import DataIncomplete
from overpy import RelationNode, RelationWay, RelationRelation
from shapely.geometry import Point as shPoint
from shapely.geometry import Polygon as shPolygon
# from swissknife.log import timeLoggerDecorator

class MPolyRolesManager(object):
    """docstring for MPolyRolesManager."""

    def __init__(self):
        super(MPolyRolesManager, self).__init__()
        self.members = {
            "inner": [],
            "outer": []
        }
        self.geoms = {}
        self.relation = {}

    def __iter__(self):
        for o,ii in self.relation.items():
            yield o, ii

    def set_member_as(self, role, way):
        try:
            self.members[role].append(way)
        except KeyError:
            pass
        else:
            self.geoms[way.id] = shPolygon([list(map(float, (n.lon,n.lat,))) for n in way.nodes])

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):

        # ~~ TODO ~~
        # WARNING! Islands within a hole are not considered at the moment.
        # Reference: http://wiki.openstreetmap.org/wiki/Relation:multipolygon
        for inner in self.members["inner"]:
            for outer in self.members["outer"]:
                if self.geoms[inner.id].within(self.geoms[outer.id]):
                    try:
                        o = self.relation[outer.id]
                    except KeyError:
                        self.relation[outer.id] = [inner.id]
                    else:
                        o.append(inner.id)


class __CommonMethods__(object):
    """docstring for __Base__."""

    def _save_way(self, info_id, way):

        if self.db.way_node(info_id=info_id) is None:
            return self.db.way_node.bulk_insert((dict(
                info_id = info_id,
                node_id =  self.cache["nodes"][node.id]["node_id"],
                sorting = sorting,
            ) for sorting, node in enumerate(way.nodes)))
        else:
            res = self.db(self.db.way_node.info_id==info_id).select(
                orderby = self.db.way_node.sorting
            )
            for srt,_dd in enumerate(zip_longest(res, way.nodes)):
                rec,node = _dd
                if node is None:
                    rec.delete_record()
                elif rec is None:
                    self.db.way_node.insert(
                        info_id = info_id,
                        node_id = self.cache["nodes"][node.id]["node_id"],
                        sorting = srt
                    )
                elif rec.node_id==self.cache["nodes"][node.id]["node_id"]:
                    # It just fix wrong values that would not exist.
                    if rec.sorting!=srt:
                        rec.update_record(sorting = srt)
                else:
                    rec.update_record(
                        node_id = self.cache["nodes"][node.id]["node_id"],
                        sorting = srt
                    )

    def parseNode(self, node):

        try:
            self.cache["nodes"][node.id]
        except KeyError:
            info_id = self._save_info(node.id,
                tags = node.tags,
                gtype = "node",
                attributes = node.attributes
            )
            coordinates = map(float, (node.lon, node.lat,))
            node_id = self._save_node(info_id, *coordinates)
            out = self.cache["nodes"][node.id] = {"id": info_id, "node_id": node_id}
            return out

    def parseWay(self, way):

        try:
            self.cache["ways"][way.id]
        except KeyError:
            info_id = self._save_info(
                way.id,
                tags = way.tags,
                gtype = "way",
                attributes = way.attributes
            )
            self._save_way(info_id, way)
            out = self.cache["ways"][way.id] = {"id": info_id}
            return out


class Parser(BaseParser, __CommonMethods__):
    """docstring for Parser."""

    def __init__(self, db, **kwargs):
        """
        db @DAL : The database in wich info, node, way_node and relation tables are defined.
        """
        super(Parser, self).__init__(db, source_name='osm')
        # Nodes and way cached by their source id
        self.cache = {"nodes": {}, "ways": {}, "relations": {}}

    def parseRelation(self, relation):
        """

        A relation is a group of elements. To be more exact it is one of the core
        data elements that consists of one or more tags and also an ordered list
        of one or more nodes, ways and/or relations as members which is used to
        define logical or geographic relationships between other elements.
        A member of a relation can optionally have a role which describes the part
        that a particular feature plays within a relation.

        References:
            * https://wiki.openstreetmap.org/wiki/Relation
        """

        info_id = self._save_info(relation.id,
            tags = relation.tags,
            attributes = relation.attributes,
            gtype = "relation"
        )

        _relations = self.db(self.db.relation.info_id==info_id)

        if _relations.select("'0'", limitby=(0,1,), cacheable=True).first() is None:

            for member in relation.members:
                if isinstance(member, RelationWay):
                    try:
                        member_id = self.cache["ways"][member.ref]["id"]
                    except KeyError:
                        continue
                elif isinstance(member, RelationNode):
                    try:
                        member_id = self.cache["nodes"][member.ref]["id"]
                    except KeyError:
                        continue
                elif isinstance(member, RelationRelation):
                    try:
                        member_id = self.cache["relations"][member.ref]["id"]
                    except KeyError:
                        try:
                            _member = member._result.get_relation(
                                member.ref, resolve_missing=False
                            )
                        except DataIncomplete:
                            continue
                        else:
                            member_id = self.parseRelation(_member)["id"]
                else:
                    raise WTF()

                self.db.relation.insert(info_id=info_id, member_id=member_id, role=member.role)
            return {"id": info_id}

        else:

            # import pdb; pdb.set_trace()
            # relations = _relations(self.db.relation.member_id==self.db.info.id).select(
            #     left = self.db.relation.on(
            #         (self.db.relation.info_id == self.db.way_node.info_id) | \
            #         (self.db.relation.info_id == self.db.node.info_id)
            #     )
            # )

            # mew_members_ref = {m.ref: m for m in relation.members}

            for row in _relations(
                (self.db.relation.member_id==self.db.info.id) & \
                ~self.db.info.source_id.belongs(list(map(lambda m: str(m.ref), relation.members)))
            ).select(self.db.relation.ALL):
                row.delete_record()

            for member in relation.members:
                if isinstance(member, RelationWay):
                    try:
                        member_id = self.cache["ways"][member.ref]["id"]
                    except KeyError:
                        continue
                elif isinstance(member, RelationNode):
                    try:
                        member_id = self.cache["nodes"][member.ref]["id"]
                    except KeyError:
                        continue
                elif isinstance(member, RelationRelation):
                    try:
                        member_id = self.cache["relations"][member.ref]["id"]
                    except KeyError:
                        try:
                            _member = member._result.get_relation(
                                member.ref, resolve_missing=False
                            )
                        except DataIncomplete:
                            continue
                        else:
                            member_id = self.parseRelation(_member)["id"]
                else:
                    raise WTF()

                self._save_relation(
                    info_id,
                    member_id,
                    role = member.role
                )
            return {"id": info_id}

    # @timeLoggerDecorator()
    def parse(self, nodes, ways, relations):
        """ Saves overpass query result into db.

        nodes     @iterable : OSM nodes iterable
        ways      @iterable : OSM ways iterable
        relations @iterable : OSM relations iterable
        """

        # ways = nodes[0]._result.ways
        # relations = nodes[0]._result.relations

        for node in nodes:
            self.parseNode(node)

        for way in ways:
            self.parseWay(way)

        for relation in relations:
            self.parseRelation(relation)


class Copier(BaseCopier, __CommonMethods__):
    """docstring for Copier."""

    def __init__(self, db, **kwargs):
        """
        db @DAL : The database in wich info, node, way_node and relation tables are defined.
        """
        super(Copier, self).__init__(db, source_name='osm')
        # Nodes and way cached by their source id
        self.cache = {
            "nodes": kwargs.get('node', {}),
            "ways": kwargs.get('way', {}),
            "relations": kwargs.get('relation', {}),
        }

    def _save_way(self, info_id, way):

        for sorting, node in enumerate(way.nodes):
            data = dict(
                info_id = info_id,
                node_id =  self.cache["nodes"][node.id]["id"],
                sorting = sorting
            )

            self._insways.writerow(data)

    def parseRelation(self, relation):
        """ """

        try:
            info_id = self.cache["relations"][relation.id]["id"]
        except KeyError:
            info_id = self._save_info(relation.id,
                tags = relation.tags,
                attributes = relation.attributes,
                gtype = "relation"
            )

        for member in relation.members:

            if isinstance(member, RelationWay):
                try:
                    member_id = self.cache["ways"][member.ref]["id"]
                except KeyError:
                    continue
            elif isinstance(member, RelationNode):
                try:
                    member_id = self.cache["nodes"][member.ref]["id"]
                except KeyError:
                    continue
            elif isinstance(member, RelationRelation):
                try:
                    member_id = self.cache["relations"][member.ref]["id"]
                except KeyError:
                    try:
                        _member = member._result.get_relation(
                            member.ref, resolve_missing=False
                        )
                    except DataIncomplete:
                        continue
                    else:
                        member_id = self.parseRelation(_member)["id"]
            else:
                raise WTF()

            self._save_relation(info_id, member_id, role=member.role)
        out = self.cache["relations"][relation.id] = {"id": info_id}
        return out

    # @timeLoggerDecorator()
    def parse(self, nodes, ways, relations):
        """
        nodes : OSM nodes
        """

        # ways = nodes[0]._result.ways
        # relations = nodes[0]._result.relations

        with BulkCopyer(self.db.relation) as self._insrelation, \
            BulkCopyer(self.db.way_node) as self._insways, \
            BulkCopyer(self.db.node) as self._insnodes, \
            BulkCopyer(self.db.info) as self._insinfo:

            # TODO: # BulkCopyer(db.filter_attribution) as self.insfilter, \

            for node in nodes:
                self.parseNode(node)

            for way in ways:
                self.parseWay(way)

            for relation in relations:
                self.parseRelation(relation)
